import numpy as np
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import textwrap


def plot2d(
    x,
    y,
    z=None,
    axis_names=["X", "Y", "Z"],
    title="2D Scatter Plot",
    desc=(
        "🔹 This plot visualizes 3D data projected onto a 2D plane if Z is provided.<br>"
        "🔹 Marker color and size represent the third dimension (Z values).<br>"
        "🔹 If Z is not provided, it simply plots X vs. Y.<br>"
        "🔹 Hover over points to see their coordinates.<br>"
        "<br> _________________________________________________________ <br>"
    ),
    html=False,
):
    """
    Generates a 2D scatter plot.

    Args:
        x (list): X-axis values.
        y (list): Y-axis values.
        z (list or None): Z values (used for color/size if provided).
        axis_names (list): Names for X, Y, and Z axes.
        title (str): Title of the plot.
        desc (str): Description text for annotations.
        html (bool): If True, return the plot as HTML instead of showing it.

    Returns:
        str or None: Returns an HTML string if `html=True`, otherwise displays the plot.
    """

    # Wrap title for better display
    title_wrap = 100
    wrapped_title = "<br>".join(textwrap.wrap(title, width=title_wrap))

    # Check if Z is provided
    use_z = z is not None

    # Create hover text showing full (X, Y, Z) coordinates if Z is provided
    if use_z:
        desc = (
            f"🔹 This plot visualizes {axis_names[0]}, {axis_names[1]}, and {axis_names[2]} data projected onto a 2D plane.<br>"
            f"🔹 Marker color and size represent the third dimension ({{axis_names[2]}} values).<br>"
            "🔹 Hover over points to see their coordinates.<br>"
            "<br> _________________________________________________________ <br>"
        )
        hover_text = [f"({x[i]}, {y[i]}, {z[i]})" for i in range(len(x))]
    else:
        desc = (
            f"🔹 This plot visualizes {axis_names[0]}, and {axis_names[1]} data projected onto a 2D plane.<br>"
            "🔹 Hover over points to see their coordinates.<br>"
            "<br> _________________________________________________________ <br>"
        )
        hover_text = [f"({x[i]}, {y[i]})" for i in range(len(x))]

    # Create figure
    fig = go.Figure()

    # Add 2D scatter plot
    marker_dict = {
        "size": (
            10
            if not use_z
            else [10 + (val - min(z)) / (max(z) - min(z) + 1) * 20 for val in z]
        ),
        "color": "blue" if not use_z else z,  # Default color if Z is not provided
        "colorscale": "Viridis" if use_z else None,
        "showscale": use_z,
        "colorbar": dict(title=f"{axis_names[2]} Values") if use_z else None,
    }

    fig.add_trace(
        go.Scatter(
            x=x,
            y=y,
            mode="markers",
            marker=marker_dict,
            text=hover_text,
            hoverinfo="text",
            name="Pareto Front" if use_z else "Data Points",
        )
    )

    # Update layout
    fig.update_layout(
        title=dict(
            text=wrapped_title,
            font=dict(size=18),
        ),
        xaxis_title=f"X: {axis_names[0]}",
        yaxis_title=f"Y: {axis_names[1]}",
        annotations=[
            go.layout.Annotation(
                showarrow=False,
                text=desc,
                x=1.02,  # Position text to the right
                y=1.02,
                xref="paper",
                yref="paper",
                font=dict(size=14, color="black"),
                bgcolor="rgba(255, 255, 255, 0.85)",
                bordercolor="black",
                borderwidth=2,
                borderpad=10,
                align="left",
            )
        ],
        width=300,  # Set plot width
        height=300,  # Set plot height
    )

    #  Return HTML if requested
    if html:
        return fig.to_html(
            full_html=False, include_plotlyjs=True
        )  # Return HTML representation of plot
    else:
        fig.show()  # Display the plot normally


def plot3d(
    x=[1, 2, 3, 4, 5],
    y=[2, 3, 1, 4, 5],
    z=[3, 1, 2, 5, 4],
    theta=[1, 2, 3, 4, 5],
    criteria="all",
    axis_names=["X", "Y", "Z"],
    title="3D Scatter Plot of Pareto front.",
    desc=(
        "🔍Hover over points to see info about associated Theta.<br>"
        "⚖️ Points show different loss trade-offs<br>"
        "🧩 Theta can be used to update the model<br>"
    ),
    html=False,
):
    """
    Generates a 3D scatter plot of Pareto front.

    Args:
        x (list): X-axis values.
        y (list): Y-axis values.
        z (list): Z-axis values.
        theta (list): Theta values.
        criteria (str): Criteria for selection.
        axis_names (list): List of axis names [X, Y, Z].
        title (str): Title of the plot.
        desc (str): Description text to be added.
        html (bool): If True, return the plot as HTML instead of showing it.

    Returns:
        str or None: Returns an HTML string if `html=True`, otherwise displays the plot.
    """

    # Wrap title for better display
    title_wrap = 100
    wrapped_title = "<br>".join(textwrap.wrap(title, width=title_wrap))

    # Hidden data for hover text
    hidden_data = [f"Theta: {theta[i]}" for i in range(len(theta))]

    # Create figure
    fig = make_subplots(rows=1, cols=1, specs=[[{"type": "scatter3d"}]])

    # Color values
    color_values = np.array(x) + np.array(y) + np.array(z)
    color_label = "Sum of losses (darker better)"

    # Add 3D scatter plot
    fig.add_trace(
        go.Scatter3d(
            x=x,
            y=y,
            z=z,
            mode="markers",
            marker=dict(
                size=8,
                color=color_values,  # Color based on the chosen attribute
                colorscale="Viridis",  # Try 'Plasma', 'Cividis', 'Inferno', etc.
                showscale=True,
                colorbar=dict(title=color_label),
            ),
            name="Points",
            text=hidden_data,
            customdata=hidden_data,
        ),
        row=1,
        col=1,
    )

    # Update layout
    fig.update_layout(
        title=dict(
            text=f"{criteria}-objectives<br>{wrapped_title}",
            font=dict(size=12),
        ),
        scene=dict(
            xaxis_title=f"X: {axis_names[0]}",
            yaxis_title=f"Y: {axis_names[1]}",
            zaxis_title=f"Z: {axis_names[2]}",
        ),
        annotations=[
            go.layout.Annotation(
                showarrow=False,
                text=desc,
                x=1.02,  # text to the right outside the plot
                y=1.02,
                xref="paper",
                yref="paper",
                font=dict(size=10, color="black"),
                bgcolor="rgba(255, 255, 255, 0.85)",
                bordercolor="black",
                borderwidth=2,
                borderpad=10,
                align="left",
            )
        ],
        width=800,  # width of the plot
        height=400,
    )

    # **NEW FEATURE**: Return HTML if requested
    if html:
        return fig.to_html(
            full_html=True, include_plotlyjs=True
        )  # Return HTML representation of plot
    else:
        fig.show()  # Display the plot normally


# import plotly.graph_objects as go
# import numpy as np
# import textwrap
# from plotly.subplots import make_subplots

# import plotly.graph_objects as go
# import numpy as np
# import textwrap
import plotly.express as px


def plot_spider(
    objectives,
    theta=None,
    criteria="all",
    axis_names=None,
    title="Multi-objective Line Chart",
    desc=(
        "🔍Hover over lines to see info about associated Theta.<br>"
        "⚖️ Lines show different loss trade-offs<br>"
        "🧩 Theta can be used to update the model<br>"
    ),
    html=False,
):
    """
    Plots a multi-objective line chart with true Viridis coloring based on sum of objectives.
    """

    num_objectives = len(objectives[0])
    num_configs = len(objectives)
    axis_names = axis_names or [f"Objective {i+1}" for i in range(num_objectives)]

    # Compute sum of objectives and normalize
    color_values = np.array([sum(obj) for obj in objectives])
    min_val, max_val = color_values.min(), color_values.max()
    norm_vals = (color_values - min_val) / (max_val - min_val + 1e-9)

    # Get Viridis colors
    viridis_colors = px.colors.sequential.Viridis
    color_map = [px.colors.sample_colorscale(viridis_colors, [v])[0] for v in norm_vals]

    fig = go.Figure()

    # Add each configuration as a line
    for i, obj_values in enumerate(objectives):
        fig.add_trace(
            go.Scatter(
                x=axis_names,
                y=obj_values,
                mode="lines+markers",
                name=str(theta[i]),
                line=dict(color=color_map[i], width=2),
                marker=dict(size=8, color=color_map[i]),
                hovertemplate=(
                    f"<b>{theta[i]}</b><br>"
                    + "<br>".join(
                        [
                            f"{axis_names[j]}: {obj_values[j]:.2f}"
                            for j in range(num_objectives)
                        ]
                    )
                    + f"<br>Sum: {color_values[i]:.2f}<extra></extra>"
                ),
            )
        )

    # Build custom colorbar manually
    colorbar_trace = go.Scatter(
        x=[None],
        y=[None],
        mode="markers",
        marker=dict(
            colorscale="Viridis",
            cmin=min_val,
            cmax=max_val,
            color=[min_val, max_val],
            colorbar=dict(title="Sum of losses (darker = better)"),
            showscale=True,
        ),
        hoverinfo="none",
    )
    fig.add_trace(colorbar_trace)

    # Wrapped title
    title_wrap = 100
    wrapped_title = "<br>".join(textwrap.wrap(title, width=title_wrap))

    # Layout
    fig.update_layout(
        title=dict(
            text=f"{criteria}-objectives<br>{wrapped_title}",
            font=dict(size=12),
            y=0.95,  # Move title higher (default is ~0.9)
        ),
        xaxis_title="Objectives",
        yaxis_title="Values",
        plot_bgcolor="#F5E8D8",
        paper_bgcolor="#F5E8D8",  
        font=dict(size=15,color="#2B2B2B"),
        annotations=[
            go.layout.Annotation(
                showarrow=False,
                text=desc,
                x=0,
                y=1.35,  # Keep it above plot
                xref="paper",
                yref="paper",
                font=dict(size=15, color="#2B2B2B"),
                bgcolor= "#EBDDCB",  #"rgba(112,128,144,1)",
                bordercolor="black",
                borderwidth=2,
                borderpad=10,
                align="left",
            )
        ],
        margin=dict(t=160),  # Adds space at the top for annotation + title
    )

    fig.update_layout(showlegend=False)

    if html:
        return fig.to_html(full_html=True, include_plotlyjs=True)
    else:
        fig.show()



import os
from dotenv import load_dotenv

load_dotenv(os.path.expanduser("~/.mmmfair_env"))

def ensure_groq_key():
    if "GROQ_API_KEY" not in os.environ:
        print("\n🔐 MMM-Fair requires an Groq API key for LLM-powered chat features.")
        print("You can get one at https://console.groq.com/keys")
        user_key = input("Please enter your Groq API key: ").strip()
        os.environ["GROQ_API_KEY"] = user_key

        save = input("Would you like to save this key for future use? (y/N): ").strip().lower()
        if save == "y":
            env_path = os.path.expanduser("~/.mmmfair_env")
            with open(env_path, "w") as f:
                f.write(f"GROQ_API_KEY={user_key}\n")
            print(f"✅ API key saved to {env_path}")
        else:
            print("🔐 Key loaded only for this session.")
    else:
        print("🔑 Found existing Groq API key in environment.")

def get_groq_llm():
    llm=None
    try:
        from langchain_groq import ChatGroq
        llm = ChatGroq(model="gemma2-9b-it", temperature=0.1, api_key=os.environ["GROQ_API_KEY"])
    except ImportError:
        print("🔕 GROQ features are disabled. Install with: pip install mmm-fair[llm-groq]")
    except Exception as e:
        print(f"⚠️ Optional GROQ setup skipped: {e}")
    return llm
        


    
    


# Author: Arjun Roy (arjun.roy@unibw.de, arjunroyihrpa@gmail.com) https://orcid.org/0000-0002-4279-9442 
# Apache License Version 2.0
import textwrap
import numpy as np
from time import time
from functools import partial
from numbers import Integral, Real
import itertools
from fairbench import v2 as fb
from tqdm import tqdm 
##sklearn ensemble
from sklearn.ensemble import HistGradientBoostingClassifier
from sklearn.ensemble._hist_gradient_boosting.grower import TreeGrower
from sklearn.ensemble._hist_gradient_boosting._gradient_boosting import _update_raw_predictions
from sklearn.ensemble._hist_gradient_boosting.binning import _BinMapper
from sklearn.ensemble._hist_gradient_boosting.common import G_H_DTYPE, X_DTYPE, Y_DTYPE
from sklearn._loss.loss import BaseLoss

#sklearn data processing and metrics
from sklearn.metrics import accuracy_score, confusion_matrix, balanced_accuracy_score
from sklearn.compose import ColumnTransformer
from sklearn.metrics import check_scoring
from sklearn.metrics._scorer import _SCORERS
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import FunctionTransformer, LabelEncoder, OrdinalEncoder

#sklearn utils
from sklearn.utils import check_random_state, compute_sample_weight, resample
from sklearn.utils._missing import is_scalar_nan
from sklearn.utils._openmp_helpers import _openmp_effective_n_threads
from sklearn.utils._param_validation import Interval, RealNotInt, StrOptions
from sklearn.utils.multiclass import check_classification_targets
from sklearn.utils import check_random_state, compute_sample_weight, resample, compute_class_weight
from sklearn.utils.validation import (
    _check_monotonic_cst,
    _check_sample_weight,
    _check_y,
    _is_pandas_df,
    check_array,
    check_consistent_length,
    check_is_fitted,
    validate_data,
)

from copy import deepcopy

from sklearn._loss import HalfBinomialLoss
from pymoo.mcdm.pseudo_weights import PseudoWeights
from typing import Literal

__all__ = ["MMM_Fair_GradientBoostedClassifier"]


def logistic_chain_rule(prob):
    """
    d(prob)/d(raw_pred) = prob * (1 - prob).
    """
    return prob * (1.0 - prob)


class MultiFairBinomialLoss(HalfBinomialLoss):
    """
    A multi-attribute fairness-aware binomial loss for binary classification.
    We combine multiple fairness penalties via a smooth max operator,
    approximating max across attributes in a differentiable manner.

    Each attribute is a binary array (0=protected,1=non-protected).
    We do e.g. 'DP' for each attribute => difference in predicted positives,
    then combine them via smoothmax with parameter 'beta'.

    Parameters
    ----------
    saIndex : array of protected values shape (n_samples,n_protected)
        For each protected attribute i, a value of 0/1 indicating group membership.
    alpha_fair : float, default=0.1
        Weight of the fairness penalty relative to logistic loss.
    beta : float, default=10.0
        Smoothmax parameter. Larger means closer to actual max but less smooth.
    fairness_type : str in {"DP","EP","EO","TPR","FPR"}, default="DP"
        Which measure to compute for each attribute.
        For multiple attributes, we do the same measure but you can adapt.
    sample_weight : array-like, optional
        For base logistic. 
    """

    def __init__(
        self,
        saIndex,
        saValue=None,
        alpha_fair=0.1,
        beta=10.0,
        fairness_type="DP",
        sample_weight=None,
    ):
        super().__init__(sample_weight=sample_weight)
        # store protected attributes as a list of arrays
        # each array shape (n_samples,)
        self.saIndex = saIndex
        if saValue is not None:
            self.saValue = saValue
        else:
            self.saValue = {}
            for j in range(self.saIndex.shape[-1]):
                self.saValue[str(j)]=0
        self.alpha_fair = alpha_fair
        self.beta = beta
        self.fairness_type = fairness_type

    def _compute_single_fairness(self, y_true, p, s):
        """
        Compute a single-attribute fairness measure:
        DP, EP, EO, TPR, FPR (toy version with squared difference).
        
        s is shape (n_samples,), 0=protected,1=non-protected
        p is predicted prob of pos class, shape (n_samples,).
        y_true is in {0,1}.
        Return:
          scalar fairness measure f_j
          array partial derivative df_j/d p[i] in shape (n_samples,).
        """
        idx0 = (s == 0)
        idx1 = (s == 1)
        n0 = max(1, idx0.sum())
        n1 = max(1, idx1.sum())

        if self.fairness_type.lower() == "dp":
            # difference in mean predicted positives
            mean0 = p[idx0].mean() if np.any(idx0) else 0.0
            mean1 = p[idx1].mean() if np.any(idx1) else 0.0
            diff = (mean1 - mean0)
            # penalty = diff^2
            penalty = diff*diff

            # derivative wrt each sample's p[i]
            grad = np.zeros_like(p)
            factor = 2.0 * diff
            for i in range(len(p)):
                if s[i] == 1:
                    chain = factor*(1./n1)
                else:
                    chain = factor*(-1./n0)
                grad[i] = chain * logistic_chain_rule(p[i])
            return penalty, grad

        elif self.fairness_type.lower() == "ep" or self.fairness_type.lower() == "tpr":
            # difference in TPR => subset y=1
            idx_pos = (y_true == 1)
            idx0p = idx0 & idx_pos
            idx1p = idx1 & idx_pos
            n0p = max(1, idx0p.sum())
            n1p = max(1, idx1p.sum())
            mean0p = p[idx0p].mean() if np.any(idx0p) else 0.
            mean1p = p[idx1p].mean() if np.any(idx1p) else 0.
            diff = mean1p - mean0p
            penalty = diff*diff

            grad = np.zeros_like(p)
            factor = 2.*diff
            for i in range(len(p)):
                if idx_pos[i]:
                    if s[i] == 1:
                        chain = factor*(1./n1p)
                    else:
                        chain = factor*(-1./n0p)
                    grad[i] = chain * logistic_chain_rule(p[i])
            return penalty, grad

        elif self.fairness_type.lower() == "fpr":
            # difference in FPR => subset y=0
            idx_neg = (y_true == 0)
            idx0n = idx0 & idx_neg
            idx1n = idx1 & idx_neg
            n0n = max(1, idx0n.sum())
            n1n = max(1, idx1n.sum())
            mean0n = (1.-p[idx0n]).mean() if np.any(idx0n) else 0.
            mean1n = (1.-p[idx1n]).mean() if np.any(idx1n) else 0.
            diff = mean1n - mean0n
            penalty = diff*diff

            grad = np.zeros_like(p)
            factor = 2.*diff
            for i in range(len(p)):
                if idx_neg[i]:
                    if s[i] == 1:
                        chain = factor*(1./n1n)*(-1.)
                    else:
                        chain = factor*(-1./n0n)*(-1.)
                    grad[i] = chain * logistic_chain_rule(p[i])
            return penalty, grad

        elif self.fairness_type.lower() == "eo":
            # TPR difference + TNR difference => for y=1, y=0
            # define penalty = (diff_tpr^2 + diff_tnr^2).
            idx_pos = (y_true == 1)
            idx_neg = (y_true == 0)
            idx1p = idx1 & idx_pos
            idx0p = idx0 & idx_pos
            idx1n = idx1 & idx_neg
            idx0n = idx0 & idx_neg

            n0p = max(1, idx0p.sum())
            n1p = max(1, idx1p.sum())
            n0n = max(1, idx0n.sum())
            n1n = max(1, idx1n.sum())

            mean0p = p[idx0p].mean() if np.any(idx0p) else 0.
            mean1p = p[idx1p].mean() if np.any(idx1p) else 0.
            diff_tpr = (mean1p - mean0p)

            # TNR => predicted negative => 1-p
            mean0n = (1.-p[idx0n]).mean() if np.any(idx0n) else 0.
            mean1n = (1.-p[idx1n]).mean() if np.any(idx1n) else 0.
            diff_tnr = (mean1n - mean0n)

            penalty = diff_tpr*diff_tpr + diff_tnr*diff_tnr

            grad = np.zeros_like(p)
            # partial derivative wrt TPR part
            factor_tpr = 2.*diff_tpr
            # partial derivative wrt TNR part
            factor_tnr = 2.*diff_tnr

            for i in range(len(p)):
                chain = 0.
                if y_true[i] == 1:
                    # TPR
                    if s[i] == 1:
                        chain += factor_tpr*(1./n1p)
                    else:
                        chain += factor_tpr*(-1./n0p)
                    # chain => derivative wrt p => multiply logistic_chain_rule
                    chain *= logistic_chain_rule(p[i])
                else:
                    # TNR => partial wrt p => negative sign
                    if s[i] == 1:
                        chain += factor_tnr*(1./n1n)*(-1.)
                    else:
                        chain += factor_tnr*(-1./n0n)*(-1.)
                    chain *= logistic_chain_rule(p[i])
                grad[i] = chain
            return penalty, grad

        else:
            # no fairness
            return 0., np.zeros_like(p)

    def _fairness_smoothmax(self, y_true, raw_pred):
        """
        For each attribute j:
          f_j, grad_j
        Then combine via softmax with param beta => smoothmax of f_j.

        Return total_penalty (scalar), grad_penalty (array shape=(n_samples,)) 
        """
        p = 1.0 / (1.0 + np.exp(-raw_pred))

        # For each attribute j => compute penalty_j, grad_j
        penalty_list = []
        grad_list = []
        for j in range(self.saIndex.shape[-1]):
            s_j=self.saIndex[:,j]
            pen_j, grad_j = self._compute_single_fairness(y_true, p, s_j)
            penalty_list.append(pen_j)
            grad_list.append(grad_j)
        
        # penalty_list => shape (m,) for m attributes
        # we define smoothmax( f_1, ..., f_m ) with param self.beta
        # => c = (1/beta)*log( sum( exp(beta * f_j) ) )
        # partial derivative wrt f_j => softmax weighting
        # partial wrt raw_pred => sum_j( w_j * d f_j/d raw_pred )
        f_arr = np.array(penalty_list, dtype=float)  # shape (m,)
        # avoid overflow
        max_f = np.max(f_arr)
        exps = np.exp(self.beta * (f_arr - max_f))
        sum_exps = np.sum(exps)
        c = (1.0/self.beta)*( np.log(sum_exps) ) + max_f  # shift back

        # derivative of c wrt f_j => w_j = exp( beta*(f_j - max_f) ) / sum_exps
        w = exps / sum_exps  # shape (m,)

        # partial derivative wrt each sample i => sum_j( w_j * grad_j[i] )
        # Combine all attributes' partial derivatives
        grad_c = np.zeros_like(grad_list[0])
        for j in range(self.saIndex.shape[-1]):
            grad_c += w[j] * grad_list[j]
        
        return c, grad_c

    def _fairness_penalty_grad(self, y_true, raw_pred):
        """
        The single-sample partial derivative of the multi-attribute
        fairness penalty. We do a smoothmax across attributes.

        Return grad_fair shape (n_samples,).
        """
        total_penalty, grad_penalty = self._fairness_smoothmax(y_true, raw_pred)
        return total_penalty, grad_penalty

    def gradient(self, y_true, raw_prediction, sample_weight, gradient_out, n_threads=1):
        """
        Overriding gradient: logistic + alpha_fair * fairness derivative
        ignoring hessian. scikit-learn calls gradient() if self.constant_hessian = True.
        """
        # call base logistic gradient
        super().gradient(y_true, raw_prediction, sample_weight, gradient_out, n_threads)
        # compute multi-attribute fairness
        _, grad_fair = self._fairness_penalty_grad(y_true, raw_prediction)
        gradient_out += self.alpha_fair * grad_fair

        return gradient_out

    def gradient_hessian(
        self,
        y_true,
        raw_prediction,
        sample_weight,
        gradient_out,
        hessian_out,
        n_threads=1
    ):
        """
        Overriding gradient & hessian. We'll do logistic for hessian, and
        add alpha_fair * fairness to gradient.
        """

        super().gradient_hessian(
            y_true, raw_prediction, sample_weight, gradient_out, hessian_out, n_threads
        )

        _, grad_fair = self._fairness_penalty_grad(y_true, raw_prediction)

        if len(gradient_out.shape)<len(grad_fair.shape):
            grad_fair = grad_fair.ravel()

        mean_log = np.mean(np.abs(gradient_out))
        mean_fair = np.mean(np.abs(grad_fair))
        ratio = mean_log / (mean_fair + 1e-7) ##slack to avoid division by zero
        gradient_out += self.alpha_fair * grad_fair*ratio

        return gradient_out,hessian_out



class MMM_Fair_GradientBoostedClassifier(HistGradientBoostingClassifier):
    def __init__(
        self,
        saIndex=None,
        saValue=None,
        constraint='DP',
        alpha=0.1,
        store_iter_metrics=True,
        preference=None,
        # plus any standard HistGradientBoostingClassifier params
        early_stopping="auto",
        validation_fraction=0.1,
        max_iter=100,
        random_state=None,
        # ...
        **kwargs
    ):

        if constraint is None:
            # your default or fallback
            super().__init__(
                loss="log_loss",
                early_stopping=early_stopping,
                validation_fraction=validation_fraction,
                max_iter=max_iter,
                random_state=random_state,
                **kwargs
            )
            self.saValue = saValue
            self.saIndex = saIndex
        else:
            if saIndex is None:
                raise ValueError("Protected features cannot be None for fairness intervention")
            else:
                self.saIndex = saIndex
                if saValue is not None:
                    self.saValue = saValue
                else:
                    self.saValue = {}
                    for j in range(self.saIndex.shape[-1]):
                        self.saValue[str(j)]=0
                self.sensitives=list(self.saValue.keys())
            super().__init__(
                loss=MultiFairBinomialLoss(
                            saIndex=self.saIndex,
                            saValue=self.saValue,
                            alpha_fair=alpha,
                            beta=10.0,
                            fairness_type=constraint,  # or "EP","EO", "TPR", "FPR"
                        ),
                early_stopping=early_stopping,
                validation_fraction=validation_fraction,
                max_iter=max_iter,
                random_state=random_state,
                **kwargs
            )
        self.alpha=alpha
        self.constraint=constraint

        if self.saValue is None or len(self.saValue)!=self.saIndex.shape[-1]:
            if saIndex is not None:
                self.saValue = {}
                for j in range(self.saIndex.shape[-1]):
                        self.saValue[str(j)]=0
                self.sensitives=list(self.saValue.keys())
            else:
                self.sensitives=[]
        self.store_iter_metrics = store_iter_metrics
        self.acc_loss_ = []
        self.bal_loss_ = []
        self.mmm_loss_ = []
        self.all_estimators=[]
        self.ob=[]
        self.feat_obs=[]
        self.fairobs=[]
        self.theta=-1
        self.pseudo=None
        self.pareto=False
        self.PF=None
        self.preference=preference
        

    def fit(self, X, y, sample_weight=None):
        """
        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            The input samples.
    
        y : array-like of shape (n_samples,)
            Target values.
    
        sample_weight : array-like of shape (n_samples,) default=None
            Weights of training data.
    
            
    
        Returns
        -------
        self : object
            Fitted estimator.
        """
        fit_start_time = time()
        acc_find_split_time = 0.0  # time spent finding the best splits
        acc_apply_split_time = 0.0  # time spent splitting nodes
        acc_compute_hist_time = 0.0  # time spent computing histograms
        # time spent predicting X for gradient and hessians update
        acc_prediction_time = 0.0
        X, known_categories = self._preprocess_X(X, reset=True)
        y = _check_y(y, estimator=self)
        y = self._encode_y(y)
        check_consistent_length(X, y)
        if self.saIndex is None:
            self.saIndex = np.random.randint(2, size=(len(X),1))
            print(f"Caution!! saIndex not provided. Model initiated with random protected values, sample: {self.saIndex[:50,0]}")
            
        check_consistent_length(X, self.saIndex)
        # Do not create unit sample weights by default to later skip some
        # computation
        
        if sample_weight is not None:
            sample_weight = _check_sample_weight(sample_weight, X, dtype=np.float64)
            # TODO: remove when PDP supports sample weights
            self._fitted_with_sw = True
    
        sample_weight = self._finalize_sample_weight(sample_weight, y)
    
        rng = check_random_state(self.random_state)
    
        # When warm starting, we want to reuse the same seed that was used
        # the first time fit was called (e.g. train/val split).
        # For feature subsampling, we want to continue with the rng we started with.
        if not self.warm_start or not self._is_fitted():
            self._random_seed = rng.randint(np.iinfo(np.uint32).max, dtype="u8")
            feature_subsample_seed = rng.randint(np.iinfo(np.uint32).max, dtype="u8")
            self._feature_subsample_rng = np.random.default_rng(feature_subsample_seed)
    
        self._validate_parameters()
        monotonic_cst = _check_monotonic_cst(self, self.monotonic_cst)
        # _preprocess_X places the categorical features at the beginning,
        # change the order of monotonic_cst accordingly
        if self.is_categorical_ is not None:
            monotonic_cst_remapped = np.concatenate(
                (
                    monotonic_cst[self.is_categorical_],
                    monotonic_cst[~self.is_categorical_],
                )
            )
        else:
            monotonic_cst_remapped = monotonic_cst
    
        # used for validation in predict
        n_samples, self._n_features = X.shape
    
        # Encode constraints into a list of sets of features indices (integers).
        interaction_cst = self._check_interaction_cst(self._n_features)
        self._in_fit = True
    
        # `_openmp_effective_n_threads` is used to take cgroups CPU quotes
        # into account when determine the maximum number of threads to use.
        n_threads = _openmp_effective_n_threads()
    
    
        if isinstance(self.loss, str):
            self._loss = self._get_loss(sample_weight=sample_weight)
        elif isinstance(self.loss, BaseLoss):
            self._loss = self.loss
    
        if self.early_stopping == "auto":
            self.do_early_stopping_ = n_samples > 10000
        else:
            self.do_early_stopping_ = self.early_stopping
    
        # create validation data if needed
        self._use_validation_data = self.validation_fraction is not None
        if self.do_early_stopping_ and self._use_validation_data:
            # stratify for classification
            # instead of checking predict_proba, loss.n_classes >= 2 would also work
            stratify = y if hasattr(self._loss, "predict_proba") else None
    
            # Save the state of the RNG for the training and validation split.
            # This is needed in order to have the same split when using
            # warm starting.
    
            if sample_weight is None:
                X_train, X_val, y_train, y_val, sa_train, sa_val = train_test_split(
                    X,
                    y,
                    self.saIndex,
                    test_size=self.validation_fraction,
                    stratify=stratify,
                    random_state=self._random_seed,
                )
                sample_weight_train = sample_weight_val = None
            else:
                # TODO: incorporate sample_weight in sampling here, as well as
                # stratify
                (
                    X_train,
                    X_val,
                    y_train,
                    y_val,
                    sa_train,
                    sa_val,
                    sample_weight_train,
                    sample_weight_val,
                ) = train_test_split(
                    X,
                    y,
                    self.saIndex,
                    sample_weight,
                    test_size=self.validation_fraction,
                    stratify=stratify,
                    random_state=self._random_seed,
                )
        else:
            X_train, y_train, sa_train, sample_weight_train = X, y, self.saIndex, sample_weight
            X_val = y_val = sa_val = sample_weight_val = None
    
    
        n_bins = self.max_bins + 1  # + 1 for missing values
        self._bin_mapper = _BinMapper(
            n_bins=n_bins,
            is_categorical=self._is_categorical_remapped,
            known_categories=known_categories,
            random_state=self._random_seed,
            n_threads=n_threads,
        )
        X_binned_train = self._bin_data(X_train, is_training_data=True)
        if X_val is not None:
            X_binned_val = self._bin_data(X_val, is_training_data=False)
        else:
            X_binned_val = None
    
        # Uses binned data to check for missing values
        has_missing_values = (
            (X_binned_train == self._bin_mapper.missing_values_bin_idx_)
            .any(axis=0)
            .astype(np.uint8)
        )
    
        if self.verbose:
            print("Fitting gradient boosted rounds:")
    
        n_samples = X_binned_train.shape[0]
        scoring_is_predefined_string = self.scoring in _SCORERS
        need_raw_predictions_val = X_binned_val is not None and (
            scoring_is_predefined_string or self.scoring == "loss"
        )
        
        # First time calling fit, or no warm start
        if not (self._is_fitted() and self.warm_start):
            # Clear random state and score attributes
            self._clear_state()
    
            # initialize raw_predictions: those are the accumulated values
            # predicted by the trees for the training data. raw_predictions has
            # shape (n_samples, n_trees_per_iteration) where
            # n_trees_per_iterations is n_classes in multiclass classification,
            # else 1.
            # self._baseline_prediction has shape (1, n_trees_per_iteration)
            self._baseline_prediction = self._loss.fit_intercept_only(
                y_true=y_train, sample_weight=sample_weight_train
            ).reshape((1, -1))
            raw_predictions = np.zeros(
                shape=(n_samples, self.n_trees_per_iteration_),
                dtype=self._baseline_prediction.dtype,
                order="F",
            )
            raw_predictions += self._baseline_prediction
    
            # predictors is a matrix (list of lists) of TreePredictor objects
            # with shape (n_iter_, n_trees_per_iteration)
            self._predictors = predictors = []
    
            # Initialize structures and attributes related to early stopping
            self._scorer = None  # set if scoring != loss
            raw_predictions_val = None  # set if use val and scoring is a string
            self.train_score_ = []
            self.validation_score_ = []
            
            
        
            if self.do_early_stopping_:
                # populate train_score and validation_score with the
                # predictions of the initial model (before the first tree)
    
                # Create raw_predictions_val for storing the raw predictions of
                # the validation data.
                if need_raw_predictions_val:
                    raw_predictions_val = np.zeros(
                        shape=(X_binned_val.shape[0], self.n_trees_per_iteration_),
                        dtype=self._baseline_prediction.dtype,
                        order="F",
                    )
    
                    raw_predictions_val += self._baseline_prediction
    
    
                ##self.scoring can be accuracy, loss, etc.
                """
                from sklearn.metrics import make_scorer, accuracy_score, mean_squared_log_error
                scoring = {
                        "accuracy": make_scorer(accuracy_score),
                        "mean_squared_log_error": make_scorer(mean_squared_log_error),
                    }
    
                """
    
                if self.scoring == "loss":
                    # we're going to compute scoring w.r.t the loss. As losses
                    # take raw predictions as input (unlike the scorers), we
                    # can optimize a bit and avoid repeating computing the
                    # predictions of the previous trees. We'll reuse
                    # raw_predictions (as it's needed for training anyway) for
                    # evaluating the training loss.
    
                    self._check_early_stopping_loss(
                        raw_predictions=raw_predictions,
                        y_train=y_train,
                        sa_train=sa_train,
                        sample_weight_train=sample_weight_train,
                        raw_predictions_val=raw_predictions_val,
                        y_val=y_val,
                        sa_val=sa_val,
                        sample_weight_val=sample_weight_val,
                        n_threads=n_threads,
                    )
                else:
                    self._scorer = check_scoring(self, self.scoring)
                    # _scorer is a callable with signature (est, X, y) and
                    # calls est.predict() or est.predict_proba() depending on
                    # its nature.
                    # Unfortunately, each call to _scorer() will compute
                    # the predictions of all the trees. So we use a subset of
                    # the training set to compute train scores.
    
                    # Compute the subsample set
                    (
                        X_binned_small_train,
                        y_small_train,
                        sample_weight_small_train,
                        indices_small_train,
                    ) = self._get_small_trainset(
                        X_binned_train,
                        y_train,
                        sample_weight_train,
                        self._random_seed,
                    )
    
                    # If the scorer is a predefined string, then we optimize
                    # the evaluation by re-using the incrementally updated raw
                    # predictions.
                    if scoring_is_predefined_string:
                        raw_predictions_small_train = raw_predictions[
                            indices_small_train
                        ]
                    else:
                        raw_predictions_small_train = None
    
                    self._check_early_stopping_scorer(
                        X_binned_small_train,
                        y_small_train,
                        sample_weight_small_train,
                        X_binned_val,
                        y_val,
                        sample_weight_val,
                        raw_predictions_small_train=raw_predictions_small_train,
                        raw_predictions_val=raw_predictions_val,
                    )
            begin_at_stage = 0
    
    
          
    
        # warm start: this is not the first time fit was called
        else:
            # Check that the maximum number of iterations is not smaller
            # than the number of iterations from the previous fit
            if self.max_iter < self.n_iter_:
                raise ValueError(
                    "max_iter=%d must be larger than or equal to "
                    "n_iter_=%d when warm_start==True" % (self.max_iter, self.n_iter_)
                )
    
            # Convert array attributes to lists
            self.train_score_ = self.train_score_.tolist()
            self.validation_score_ = self.validation_score_.tolist()
    
            # Compute raw predictions
            raw_predictions = self._raw_predict(X_binned_train, n_threads=n_threads)
            if self.do_early_stopping_ and need_raw_predictions_val:
                raw_predictions_val = self._raw_predict(
                    X_binned_val, n_threads=n_threads
                )
            else:
                raw_predictions_val = None
    
            if self.do_early_stopping_ and self.scoring != "loss":
                # Compute the subsample set
                (
                    X_binned_small_train,
                    y_small_train,
                    sample_weight_small_train,
                    indices_small_train,
                ) = self._get_small_trainset(
                    X_binned_train, y_train, sample_weight_train, self._random_seed
                )
                sa_train=self.saIndex[indices_small_train]
    
            # Get the predictors from the previous fit
            predictors = self._predictors
    
            begin_at_stage = self.n_iter_
    
        # initialize gradients and hessians (empty arrays).
        # shape = (n_samples, n_trees_per_iteration).
        gradient, hessian = self._loss.init_gradient_and_hessian(
            n_samples=n_samples, dtype=G_H_DTYPE, order="F"
        )
    
    
        progress_bar = tqdm(range(begin_at_stage, self.max_iter), desc="MMM_Fair gradboost", unit="round")
        for iteration in progress_bar:#range(begin_at_stage, self.max_iter):
            if self.verbose >= 2:
                iteration_start_time = time()
                print(
                    "[{}/{}] ".format(iteration + 1, self.max_iter), end="", flush=True
                )
    
            # Update gradients and hessians, inplace
            # Note that self._loss expects shape (n_samples,) for
            # n_trees_per_iteration = 1 else shape (n_samples, n_trees_per_iteration).
            if self._loss.constant_hessian:
                self._loss.saIndex=sa_train
                self._loss.gradient(
                    y_true=y_train,
                    raw_prediction=raw_predictions,
                    sample_weight=sample_weight_train,
                    gradient_out=gradient,
                    n_threads=n_threads,
                )
            else:
                self._loss.saIndex=sa_train
                self._loss.gradient_hessian(
                    y_true=y_train,
                    raw_prediction=raw_predictions,
                    sample_weight=sample_weight_train,
                    gradient_out=gradient,
                    hessian_out=hessian,
                    n_threads=n_threads,
                )
    
            # Append a list since there may be more than 1 predictor per iter
            predictors.append([])
    
            # 2-d views of shape (n_samples, n_trees_per_iteration_) or (n_samples, 1)
            # on gradient and hessian to simplify the loop over n_trees_per_iteration_.
            if gradient.ndim == 1:
                g_view = gradient.reshape((-1, 1))
                h_view = hessian.reshape((-1, 1))
            else:
                g_view = gradient
                h_view = hessian
    
            # Build `n_trees_per_iteration` trees.
            for k in range(self.n_trees_per_iteration_):
                grower = TreeGrower(
                    X_binned=X_binned_train,
                    gradients=g_view[:, k],
                    hessians=h_view[:, k],
                    n_bins=n_bins,
                    n_bins_non_missing=self._bin_mapper.n_bins_non_missing_,
                    has_missing_values=has_missing_values,
                    is_categorical=self._is_categorical_remapped,
                    monotonic_cst=monotonic_cst_remapped,
                    interaction_cst=interaction_cst,
                    max_leaf_nodes=self.max_leaf_nodes,
                    max_depth=self.max_depth,
                    min_samples_leaf=self.min_samples_leaf,
                    l2_regularization=self.l2_regularization,
                    feature_fraction_per_split=self.max_features,
                    rng=self._feature_subsample_rng,
                    shrinkage=self.learning_rate,
                    n_threads=n_threads,
                )
                grower.grow()
    
                acc_apply_split_time += grower.total_apply_split_time
                acc_find_split_time += grower.total_find_split_time
                acc_compute_hist_time += grower.total_compute_hist_time
    
                ### _update_leaves_values only uses loss.fit_intercept_only, so saIndex not needed
                if not self._loss.differentiable:
                    _update_leaves_values(
                        loss=self._loss,
                        grower=grower,
                        y_true=y_train,
                        raw_prediction=raw_predictions[:, k],
                        sample_weight=sample_weight_train,
                    )
    
                predictor = grower.make_predictor(
                    binning_thresholds=self._bin_mapper.bin_thresholds_
                )
                predictors[-1].append(predictor)
    
                # Update raw_predictions with the predictions of the newly
                # created tree.
                tic_pred = time()
                _update_raw_predictions(raw_predictions[:, k], grower, n_threads)
                toc_pred = time()
                acc_prediction_time += toc_pred - tic_pred
    
    
    
            should_early_stop = False
            if self.do_early_stopping_:
                # Update raw_predictions_val with the newest tree(s)
                if need_raw_predictions_val:
                    for k, pred in enumerate(self._predictors[-1]):
                        raw_predictions_val[:, k] += pred.predict_binned(
                            X_binned_val,
                            self._bin_mapper.missing_values_bin_idx_,
                            n_threads,
                        )
    
                if self.scoring == "loss":
                    should_early_stop = self._check_early_stopping_loss(
                        raw_predictions=raw_predictions,
                        y_train=y_train,
                        sa_train=sa_train,
                        sample_weight_train=sample_weight_train,
                        raw_predictions_val=raw_predictions_val,
                        y_val=y_val,
                        sa_val=sa_val,
                        sample_weight_val=sample_weight_val,
                        n_threads=n_threads,
                    )
    
                else:
                    # If the scorer is a predefined string, then we optimize the
                    # evaluation by re-using the incrementally computed raw predictions.
                    if scoring_is_predefined_string:
                        raw_predictions_small_train = raw_predictions[
                            indices_small_train
                        ]
                    else:
                        raw_predictions_small_train = None
    
                    should_early_stop = self._check_early_stopping_scorer(
                        X_binned_small_train,
                        y_small_train,
                        sample_weight_small_train,
                        X_binned_val,
                        y_val,
                        sample_weight_val,
                        raw_predictions_small_train=raw_predictions_small_train,
                        raw_predictions_val=raw_predictions_val,
                    )
                    ####hola
    
    
            if self.verbose >= 2:
                self._print_iteration_stats(iteration_start_time)
    
            # maybe we could also early stop if all the trees are stumps?
            if should_early_stop:
                break
    
        if self.verbose:
            duration = time() - fit_start_time
            n_total_leaves = sum(
                predictor.get_n_leaf_nodes()
                for predictors_at_ith_iteration in self._predictors
                for predictor in predictors_at_ith_iteration
            )
            n_predictors = sum(
                len(predictors_at_ith_iteration)
                for predictors_at_ith_iteration in self._predictors
            )
            print(
                "Fit {} trees in {:.3f} s, ({} total leaves)".format(
                    n_predictors, duration, n_total_leaves
                )
            )
            print(
                "{:<32} {:.3f}s".format(
                    "Time spent computing histograms:", acc_compute_hist_time
                )
            )
            print(
                "{:<32} {:.3f}s".format(
                    "Time spent finding best splits:", acc_find_split_time
                )
            )
            print(
                "{:<32} {:.3f}s".format(
                    "Time spent applying splits:", acc_apply_split_time
                )
            )
            print(
                "{:<32} {:.3f}s".format("Time spent predicting:", acc_prediction_time)
            )
    
        self.train_score_ = np.asarray(self.train_score_)
        self.validation_score_ = np.asarray(self.validation_score_)
        self.all_estimators=[predictor for predictor in self._predictors]
        
    
        if self.store_iter_metrics:
            X_ = check_array(X, dtype=None, accept_sparse=False)
            check_is_fitted(self)
            for ypred in self.staged_predict(X_):
                o1=1-accuracy_score(y_pred=ypred,y_true=y)
                o2=1-balanced_accuracy_score(y_pred=ypred,y_true=y).item()
                o3=[]
                for i in range(self.saIndex.shape[-1]):
                    sens = fb.categories(self.saIndex[:, i])
                    report = fb.reports.pairwise(
                        predictions=ypred,
                        labels=y,
                        sensitive=sens
                    )
                    report_dict=report.show(env=fb.export.ToDict)
                    dp=report_dict['depends'][6]['depends'][1]['value']['value']
                    ep=report_dict['depends'][6]['depends'][2]['value']['value']
                    tpr = report_dict['depends'][6]['depends'][2]['value']['value']
                    fpr = report_dict['depends'][6]['depends'][3]['value']['value']
                    eo=max(report_dict['depends'][6]['depends'][2]['value']['value'],
                                      report_dict['depends'][6]['depends'][3]['value']['value'])
                    if self.constraint.lower()=='dp':
                        o3.append(dp)
                    elif self.constraint.lower()=='ep':
                        o3.append(ep)
                    elif self.constraint.lower()=='eo':
                        o3.append(eo)
                    elif self.constraint.lower()=='tpr':
                        o3.append(tpr)
                    elif self.constraint.lower()=='fpr':
                        o3.append(fpr)
                    else:
                        o3.append(0.99)
                self.ob.append([o1, o2, max(o3)])
                self.feat_obs.append(o3) 
                self.fairobs.append([dp,ep,eo,tpr,fpr])
                
                
        self.ob = np.array(self.ob)
        self.feat_obs = np.array(self.feat_obs)
        self.fairobs = np.array(self.fairobs)
        del self._in_fit  # hard delete so we're sure it can't be used anymore
        return self
    
    
    def update_theta(
        self,
        criteria: Literal["all", "fairness","fairdefs"] = "all",
        preference=[0.33, 0.34, 0.33],
        theta=None ##for bruteforce theta update
    ):
        def is_pareto(costs, maximise=False):
            """
            :param costs: An (n_points, n_costs) array
            :maximise: boolean. True for maximising, False for minimising
            :return: A (n_points, ) boolean array, indicating whether each point is Pareto efficient
            """
            eps = 0.001
            is_efficient = np.ones(costs.shape[0], dtype=bool)
            for i, c in enumerate(costs):
                if is_efficient[i]:
                    if maximise:
                        is_efficient[is_efficient] = np.any(
                            costs[is_efficient] >= c, axis=1
                        )  # Remove dominated points
                    else:
                        is_efficient[is_efficient] = np.any(
                            costs[is_efficient] <= c, axis=1
                        )  # Remove dominated points
            return is_efficient

        if theta is None:
            #self.preference = preference
            best_theta = 0
            if criteria.lower() == "fairness":
                objective = deepcopy(self.feat_obs)
                if len(preference)!= self.feat_obs.shape[-1]:
                    preference = [1/self.feat_obs.shape[-1] for i in range(self.feat_obs.shape[-1])]
            elif criteria.lower() == "fairdefs":
                objective = deepcopy(self.fairobs)
            else:
                objective = deepcopy(self.ob)
            # objective=np.round(objective,2)
            if self.pareto == False:
                PF = {i: objective[i] for i in range(len(objective)) if np.any(objective[i]==0)==False}
                F = np.array([objective[o] for o in range(len(objective))])
                self.PF = PF
            else:
                pf = is_pareto(objective)
                PF = {i: objective[i] for i in range(len(pf)) if pf[i] == True and np.any(objective[i]==0)==False}
                F = np.array(list(PF.values()))
                self.PF = PF
    
            if self.preference is None or len(self.preference)!= objective.shape[-1]:
                weights = preference  ##Preference Weights
            else:
                weights = self.preference
    
            best_theta, pseudo_weights = PseudoWeights(weights).do(
                F, return_pseudo_weights=True
            )
    
            if self.preference == None:
                sum_W = [sum((1 - pseudo_weights[w]) * F[w]) for w in range(len(PF))]
                best_theta = sum_W.index(min(sum_W))
    
            self.theta = list(PF.keys())[best_theta] + 1
            self.pseudo = pseudo_weights
            self._predictors=self.all_estimators[:self.theta]
        else:
            self.theta =theta
            self._predictors=self.all_estimators[:self.theta]
    
    def _check_early_stopping_loss(
        self,
        raw_predictions,
        y_train,
        sa_train,
        sample_weight_train,
        raw_predictions_val,
        y_val,
        sa_val,
        sample_weight_val,
        n_threads=1,
    ):
        """Check if fitting should be early-stopped based on loss.
    
        Scores are computed on validation data or on training data.
        """
        self._loss.saIndex=sa_train
        self.train_score_.append(
            -self._loss(
                y_true=y_train,
                raw_prediction=raw_predictions,
                sample_weight=sample_weight_train,
                n_threads=n_threads,
            )
        )
    
        if self._use_validation_data:
            self._loss.saIndex=sa_val
            self.validation_score_.append(
                -self._loss(
                    y_true=y_val,
                    raw_prediction=raw_predictions_val,
                    sample_weight=sample_weight_val,
                    n_threads=n_threads,
                )
            )
            return self._should_stop(self.validation_score_)
        else:
            return self._should_stop(self.train_score_)
    
    def see_pareto(self):
        import plotly.graph_objs as go
        from plotly.subplots import make_subplots
        def plot2d(x,y,axis_names):
            pass
        def plot3d(x = [1, 2, 3, 4, 5],y = [2, 3, 1, 4, 5],z = [3, 1, 2, 5, 4],theta=[1,2,3,4,5],criteria='all',axis_names=['X','Y','Z'], title="3D Scatter Plot of Pareto front.", desc=(
                    "🔹 This plot visualizes our data in 3D.<br>"
                    "🔹 Each point's position is determined by X, Y, Z coordinates.<br>"
                    "🔹 Marker color is blue by default, but could represent categories.<br>"
                    "🔹 Hover over points to see info about the associated **theta** (ensemble pointer).<br>"
                    "🔹 Theta can be used as preferences to update the model:<br>"
                    "➡ <b>Example: model.update_theta(theta=my_preferred_theta)</b>"
            "<br> _________________________________________________________ <br> "
                ) ):
            # Create the figure
            title_wrap=100
            wrapped_title = "<br>".join(textwrap.wrap(title, width=title_wrap))
            hidden_data=[f'Theta: {theta[i]}' for i in range(len(theta))]
            fig = make_subplots(rows=1, cols=1, specs=[[{'type': 'scatter3d'}]])
            color_values = np.array(x) + np.array(y) + np.array(z)
            color_label = "Sum of losses (darker better)"
            # Add 3D scatter plot
            fig.add_trace(go.Scatter3d(x=x, y=y, z=z, mode='markers', marker=dict(size=8,
                                                color=color_values,  # Color based on the chosen attribute
                                                colorscale="Viridis",  # Try 'Plasma', 'Cividis', 'Inferno', etc.
                                                showscale=True,
                                                colorbar=dict(title=color_label),
                                            ), name='Points',text=hidden_data, customdata=hidden_data ), 
                          row=1, col=1)
            
            # Define callback function for selection event
            def update_point(trace, points, selector):
                selected_points = [(trace.x[i], trace.y[i], trace.z[i]) for i in points.point_inds]
                selected_hidden_data = [trace.customdata[i] for i in points.point_inds]
                print("Selected Points:", selected_points,selected_hidden_data)
            
            # Assign callback function to plot
            fig.data[0].on_selection(update_point)
            
            # Update layout
            fig.update_layout(title=dict(
                                    text=f"{criteria}-objectives<br>{wrapped_title}",
                                    #x=0.5,  # Center the title
                                    #y=1.2,  # Push title down slightly to prevent clipping
                                    font=dict(size=18),
                                ),
                              scene=dict(
                                        xaxis_title='X:'+axis_names[0],
                                        yaxis_title='Y:'+ axis_names[1],
                                        zaxis_title='Z:'+ axis_names[2]
                                    ),
                             annotations=[
                                            go.layout.Annotation(
                                            showarrow=False,
                                            text=desc,
                                            x=1.02,  # text to the right outside the plot
                                            y=1.02,
                                            xref="paper",
                                            yref="paper",
                                            font=dict(size=14, color="black"),
                                            bgcolor="rgba(255, 255, 255, 0.85)",  
                                            bordercolor="black",
                                            borderwidth=2,
                                            borderpad=10,
                                            align="left",
                                        )
                                                                    ],
                                    width=1300,  # width of the plot
                                    height=700 )
            
            # Show plot
            fig.show()
        
        #self.update_theta(criteria='all')
        PF=np.array([self.ob[i] for i in range(len(self.ob))])
        thetas=np.arange(len(self.ob))
        title=f"3D Scatter Plot. Showing various trade-off points between Accuracy, Balanced Accuracy, and Maximum violation of {self.constraint} fairness among protected attributes."
        plot3d(x=PF[:,0],y=PF[:,1],z=PF[:,2], theta=thetas, criteria="Multi",
               axis_names=['Acc.','Balanc. Acc', 'MMM-fair'],title=title)
        PF=np.array([self.fairobs[i] for i in range(len(self.fairobs))])
        title=f"3D Scatter Plot. Showing various trade-off points between maximum violation of Demopgraphic Parity, Equal Opportunity, and Equalized odds fairness for the given set of protected attributes."
        plot3d(x=PF[:,0],y=PF[:,1],z=PF[:,2], theta=thetas, criteria= "Multi-definitions",
               axis_names=['DP','EqOpp', 'EqOdd'],title=title)
        if len(self.sensitives)>0:
            #self.theta=None
            #self.update_theta(criteria='fairness')
            PF=np.array([self.feat_obs[i] for i in range(len(self.feat_obs))])
            title=f"3D Scatter Plot. Showing various trade-off points between violation of {self.constraint} fairness among the protected attributes {self.sensitives}."
            if PF.shape[-1]>2:
                plot3d(x=PF[:,0],y=PF[:,1],z=PF[:,2], theta=thetas,
                       criteria='Multi-attribute',axis_names=self.sensitives,title=title)
            elif PF.shape[-1]==2:
                    plot3d(x=PF[:,0],y=PF[:,1],z=np.zeros_like(PF[:,1]), theta=thetas,
                       criteria='Multi-attribute',axis_names=self.sensitives+[''],title=title)
        
            else:
                print('Not a MMM-fair')   
    
    





import onnxruntime as rt
import numpy as np
import re
import numpy as np
import zipfile

class ONNX_MMM():
    def __init__(
        self,
        models,
        alphas=None,
        classes=None,
        n_classes=None,
        theta=None,
        pareto=None,
        sensitives=None,
    ):
        self.models = models
        self.pareto = pareto
        self.alphas = alphas
        self.sensitive = sensitives
        self.classes = classes
        self.theta = theta
        self.n_classes = n_classes

    def predict(self, dataset, sensitive, theta=None):
        """assert (
            sensitive is None or len(sensitive) == 0
        ), "ONNXEnsemble can only be called with no declared sensitive attributes" """
        theta = theta if self.theta is None else self.theta
        sensitive = sensitive if self.sensitive is None else self.sensitive
        X = dataset.to_pred(sensitive)
        X = X.astype(np.float32)
        if len(self.models)>1:
            try:
                classes = self.classes[:, np.newaxis]
        
                pred = 0
                i = 0
                for estimator, alpha in zip(
                    self.models[:theta],
                    self.alphas[:theta],
                ):
                    #notify_progress(i / theta, f"Running ensemble voter {i}/{theta}")
                    session = rt.InferenceSession(estimator)
                    input_name = session.get_inputs()[0].name
                    pred += (session.run(None, {input_name: X})[0] == classes).T * alpha
                    i += 1
                #notify_progress(1, f"Completed ensemble voting")
                pred /= self.alphas[:theta].sum()
                pred[:, 0] *= -1
                preds = classes.take(pred.sum(axis=1) > 0, axis=0)
                return np.squeeze(preds, axis=1)   
            except Exception as e:
                print("ONNX predict failed:", e)
                raise ValueError(f'Inputs or params mismatch for mmm ensemble: {e}')
        else:
            try:
                model_bytes=self.models[0]
                session = rt.InferenceSession(model_bytes)
                input_name = session.get_inputs()[0].name
                outputs = session.run(None, {input_name: X})
                return outputs[0]
            except Exception as e:
                print("ONNX predict failed:", e)
                raise ValueError(f'Inputs or params mismatch: {e}')


'''
Dataset class copied from https://github.com/mammoth-eu/mammoth-commons
'''

class Dataset:
    #integration = "dsl.Dataset"

    def to_features(self, sensitive):
        raise Exception(
            f"{self.__class__.__name__} has no method to_features(sensitive)"
        )

    def format_description(self):
        if not hasattr(self, "description"):
            return ""
        dataset_desc = "<h1>Dataset</h1>"
        if isinstance(self.description, str):
            dataset_desc += self.description + "<br>"
        elif isinstance(self.description, dict):
            for key, value in self.description.items():
                dataset_desc += f"<h3>{key}</h3>" + value.replace("\n", "<br>") + "<br>"
        else:
            raise Exception("Dataset description must be a string or a dictionary.")
        return dataset_desc



import os
import time
from flask import Flask, render_template, request, session, jsonify
from flask_session import Session
import argparse
import pandas as pd
import numpy as np
import json
import plotly.graph_objects as go
from flask import send_from_directory
# Import required modules
from mmm_fair.train_and_deploy import (
    generate_reports, 
    parse_base_learner,
    parse_numeric_input,
    build_sensitives,
    train
)
from mmm_fair.deploy_utils import convert_to_onnx
from mmm_fair.viz_trade_offs import plot2d, plot3d
from mmm_fair.mmm_fair import MMM_Fair
from mmm_fair.mmm_fair_gb import MMM_Fair_GradientBoostedClassifier
from mmm_fair.data_process import data_uci, data_local  # Importing data processing module
try:
    from mmm_fair.mchat.langchain_utils import (get_langchain_agent, 
                                        summarize_html_report, 
                                        get_available_llm_providers)
    LLM_AVAILABLE = True
except ImportError:
    print("🔕 LLM features not installed. To enable, use: pip install 'mmm-fair[llm-gpt]' or similar.")
    LLM_AVAILABLE = False
    
from mmm_fair.mchat.openai_utils import get_openAI_llm
from mmm_fair.mchat.groq_utils import get_groq_llm


from mmm_fair.viz_trade_offs import plot_spider
from mmm_fair.dataset_visualization import generate_nested_pie_chart

import uuid
import plotly.io as pio
llm=None

BASE_DIR = os.path.dirname(__file__)
PLOT_DIR = os.path.join(BASE_DIR, "static")
os.makedirs(PLOT_DIR, exist_ok=True)

app = Flask(__name__,
            template_folder=os.path.join(BASE_DIR, "templates"),
            static_folder=PLOT_DIR)
app.secret_key = "SOME_SECRET"
app.config["SESSION_TYPE"] = "filesystem"
Session(app)


THETA_DIV = """ 
<div id="theta-selection">
    <h3>Select Theta for Model Update</h3>
    <input type="number" id="theta-input" min="0" placeholder="Enter Theta index">
    <button id="theta-btn">Update Model</button>

    <h3>Save Model</h3>
    <input type="text" id="save-path" placeholder="Enter directory path">
    <button id="save-btn">Save Onnx</button>
</div>
"""

# Define chat arguments
CHAT_ARGS = [
    "dataset",
    "target",
    "prots",
    "nprotgs",
    "pos_Class",
    "data_visualization",
    "classifier",
    "constraint",
    "n_learners",
    # "save_as",
    # "moo_vis"
]

default_args = {
    "classifier": "MMM_Fair_GBT",
    "dataset": "Adult",
    "target": "income",
    "pos_Class": ">50K",
    "n_learners": 100,
    "prots": ["race", "sex"],
    "nprotgs": ["White", "Male"],
    "constraint": "EO",
    "save_as": "Onnx",
    "save_path": "my_mmm_fair_model",
    "base_learner": None,
    "report_type": "table",
    "pareto": False,
    "test": 0.3,
    "early_stop": False,
    "moo_vis": True,
}


# Dictionary of datasets and their recommended features
DATASET_RECOMMENDATIONS = {
    "adult": ["race", "sex", "age"],
    "bank": ["age", "marital", "education"],
    "credit": ["X2", "X4", "X5"],
    "kdd": ["AAGE", "ASEX", "ARACE"],
    "upload_data" : [],
}

TARGET_SUGGESTIONS = {
    "adult": [
        {"value": "income", "text": "income"},
    ],
    "bank": [
        {"value": "y", "text": "y (subscription)"},
    ],
    "credit": [
        {"value": "Y", "text": "default"},
    ],
    "kdd": [
        {"value": "income", "text": "income"},
    ],
    "upload_data" : [],
}

POS_CLASS_SUGGESTIONS = {
    "adult": {
        "income": [
            {"value": ">50K", "text": ">50K"},
            {"value": "<=50K", "text": "<=50K"},
        ],
    },
    "bank": {
        "y": [{"value": "yes", "text": "yes"}, {"value": "no", "text": "no"}],
    },
    "credit": {
        "default": [
            {"value": "1", "text": "1 (defaulted)"},
            {"value": "0", "text": "0 (not defaulted)"},
        ],
    },
    "kdd": {
        "income": [
            {"value": ' 50000+.', "text": ">50K$ (high income)"},
            {"value": '-50000', "text": "<50K$ (low income)"},
        ],
    },
    "upload_data" : {},
}

# Add this dictionary to your app.py file right after DATASET_RECOMMENDATIONS
NONPROTECTED_SUGGESTIONS = {
    "adult": {
        "race": [
            {"value": "White", "text": "White"},
            {"value": "Black", "text": "Black"},
            {"value": "Asian-Pac-Islander", "text": "Asian-Pac-Islander"},
        ],
        "sex": [
            {"value": "Male", "text": "Male"},
            {"value": "Female", "text": "Female"},
        ],
        "age": [
            {"value": "18_30", "text": "18-30"},
            {"value": "30_45", "text": "30-45"},
            {"value": "45_60", "text": "45-60"},
        ],
    },
    "bank": {
        "age": [
            {"value": "18_30", "text": "18-30"},
            {"value": "30_45", "text": "30-45"},
        ],
        "marital": [
            {"value": "married", "text": "Married"},
            {"value": "single", "text": "Single"},
        ],
        "education": [
            {"value": "primary", "text": "Primary"},
            {"value": "secondary", "text": "Secondary"},
            {"value": "tertiary", "text": "Tertiary"},
        ],
    },
    "credit": {
        "X2": [
            {"value": "1", "text": "1=Male"},
            {"value": "2", "text": "2=Female"},
        ],
        "X5": [
            {"value": "18_30", "text": "age: 18-30"},
            {"value": "30_45", "text": "age: 30-45"},
        ],
        "X4": [{"value": "1", "text": "1=married"}, {"value": "2", "text": "2=single"}, {"value": "3", "text": "3=others"}],
    },
    "kdd": {
        "ARACE": [
                    {"value": " White", "text": " White"},
                    {"value": " Asian or Pacific Islander", "text": " Asian or Pacific Islander"},
                    {"value": " Amer Indian Aleut or Eskimo", "text": " Amer Indian Aleut or Eskimo"},
                    {"value": " Black", "text": " Black"},
                        {"value": " Other", "text": " Other"}
                ],
        "ASEX": [{"value": ' Male', "text": "Male"}, {"value": ' Female', "text": "Female"}],
        "AAGE": [
            {"value": "18_30", "text": "18-30"},
            {"value": "30_45", "text": "30-45"},
            {"value": "45_60", "text": "45-60"},
            {"value": "60_80", "text": "Above 60"},
        ],
        
    },
    #"upload_data" : {},
}

# Dictionary to store available features for each dataset
DATASET_FEATURES = {}

# DEFAULT_BOT_MESSAGES = [
#     {
#         "sender": "bot",
#         "text": "Hello! Welcome to MMM-Fair Chat.\n\n"
#         "I can help you train either MMM_Fair (AdaBoost style) or MMM_Fair_GBT (Gradient Boosting style).\n"
#         "Please select an option below:",
#         "options": [
#             {"value": "MMM_Fair", "text": "MMM_Fair (AdaBoost)"},
#             {"value": "MMM_Fair_GBT", "text": "MMM_Fair_GBT (Gradient Boosting)"},
#             {"value": "default", "text": "Run with default parameters"},
#         ],
#     }
# ]
DEFAULT_BOT_MESSAGES = [
    {
        "sender": "bot",
        "text": "👋 Hello! Welcome to MMM-Fair Chat.\n\n"
                "Let's get started by selecting a dataset to work with.\n"
                "You can choose from well-known public datasets or upload your own CSV file.",
        "options": [
            {"value": "Adult", "text": "🧑 Adult (UCI)"},
            {"value": "Bank", "text": "🏦 Bank Marketing (UCI)"},
            {"value": "Credit", "text": "💳 Credit Default (UCI)"},
            {"value": "kdd", "text": "🧑 Census-Income KDD (UCI)"},
            {"value": "upload_data", "text": "📁 Upload your own Data (currently supported files: '.csv'"},
        ],
    }
]
@app.route("/")
def index():
    session["llm_enabled"] = False
    session["api_enabled"] = False
    return render_template("index.html")

@app.route("/upload_data", methods=["POST"])
def upload_csv():
    try:
        uploaded_file = request.files.get("file")
        if not uploaded_file or not uploaded_file.filename.endswith(".csv"):
            return jsonify({"success": False, "error": "Only CSV files are supported."})
        
        # Read and store dataframe in session
        df = pd.read_csv(uploaded_file)
        session["data"] = data_local(df) # Save to session
        #session["dataset_full"] = df
        return jsonify({"success": True, "columns": df.columns.tolist()})
    except Exception as e:
        return jsonify({"success": False, "error": str(e)})

@app.route("/provide_api_key", methods=["POST"])
def provide_api_key():
    try:
        env_key_map = {
                "openai": "OPENAI_API_KEY",
                "chatgpt": "OPENAI_API_KEY",
                "groq": "GROQ_API_KEY",
                "groqai": "GROQ_API_KEY",
                "together": "TOGETHER_API_KEY"
            }
        data = request.json
        api_key = data.get("api_key", "").strip()
        model = data.get("model", "").strip()
        if not api_key:
            return jsonify({"success": False, "error": "No API key provided."})

        if not model:
            return jsonify({"success": False, "error": "No LLM provider specified."})
        
        env_var = env_key_map.get(model.lower())
        if not env_var:
            return jsonify({"success": False, "error": f"Unknown provider '{model}'."})
            
        os.environ[env_var] = api_key
        session["api_enabled"] = True

        return jsonify({"success": True, "message": "✅ API key accepted. LLM features are now enabled."})

    except ImportError:
        return jsonify({
            "success": False,
            "error": "LangChain/OpenAI module not installed. Install with: pip install mmm-fair[llm-gpt]"
        })
    except Exception as e:
        return jsonify({
            "success": False,
            "error": f"Unexpected error occurred: {str(e)}"
        })
    

@app.route("/start_chat", methods=["POST"])
def start_chat():
    session.clear()
    #session["chat_history"] = list(DEFAULT_BOT_MESSAGES)
    session["user_args"] = {}
    session["plot_all_url"]=''
    if session.get("llm_enabled"):      
        template=[("system","You are an assistant inside the MMM-Fair Chat."),
                  ("User", "Greet the user politely and explain that this tool can train fairness-aware models.Use the following context: {context}")]
        llm = get_openAI_llm()
        llm_agent=get_langchain_agent(llm,template)
        message = langchain_agent.invoke({"context": DEFAULT_BOT_MESSAGES[0]["text"]})
        session["chat_history"] = [{"sender": "bot", "text": message}]
    else:
        session["chat_history"] = list(DEFAULT_BOT_MESSAGES)
    return jsonify({"ok": True})


def prompt_for_llm_provider(chat_history):
    if LLM_AVAILABLE:
        available = get_available_llm_providers()
    else:
        available = LLM_AVAILABLE
    session["valid_providers"] = available
    if not available:
        chat_history.append({
            "sender": "bot",
            "text": "⚠️ No LLM providers are available. Please install with:\n"
                    "`pip install mmm-fair[llm-gpt]`, `[llm-groq]`, or `[llm-together]`"
        })
    else:
        session["llm_enabled"]=True

        available_options = [{"value": opt, "text": opt} for opt in list(available.keys())]
        
        chat_history.append({
            "sender": "bot",
            "text": f"🧠 Please select an available provider for explanation:",
            "options": available_options
        })
    return chat_history

def handle_llm_provider_selection(user_msg, chat_history):
    """Handle the LLM provider selection and set up for API key request."""
    user_selection = user_msg.lower()
    valid_providers = session.get("valid_providers")
    
    if user_selection in valid_providers:
        # Set the provider in the session
        session["llm_provider"] = user_selection
        
        # Set a flag to indicate we need an API key
        session["needs_api_key"] = True
        
        # Add a message to the chat history
        chat_history.append({
            "sender": "bot",
            "text": f"✅ Provider '{user_selection}' selected. Setting up for API key input..."
        })
        
        return chat_history, True  # Return True as second value to indicate API key needed
    else:
        chat_history.append({
            "sender": "bot",
            "text": f"⚠️ Unknown provider '{user_selection}'. Please choose one of: {', '.join(list(valid_providers.keys()))}"
        })
        
        return chat_history, False  # Return False to indicate no API key needed yet

@app.route("/ask_chat", methods=["POST"])
def ask_chat():
    chat_history = session.get("chat_history", [])
    user_args = session.get("user_args", {})
    user_args["df"] = None
    if "file" in request.files:
        user_msg = request.form.get("message", "").strip()
        selected_features = request.form.get(
            "selected_features", []
        )
        nprotgs_value = request.form.get("nprotgs_value", "")
        button_value = request.form.get("button_value", "")
        if button_value:
            user_msg = button_value
    else:
        
        user_msg = request.json.get("message", "").strip()
        button_value = request.json.get("button_value", "")
        selected_features = request.json.get(
            "selected_features", []
        )  # Get selected features if provided
    
        # Add this to handle non-protected group selection
        nprotgs_value = request.json.get("nprotgs_value", "")

    #print(f"DEBUG: Received button_value: {button_value}")
    #print("DEBUG: dataset =", user_args.get("dataset"))

    # If a button was clicked, use its value instead of the message
    if button_value:
        user_msg = button_value
        

    if session.get("mmm_classifier") is not None and session.get("last_action") == "plotted" and user_msg.lower() in ["yes", "explain", "please explain", "what do these mean?", "openai", "chatgpt", "groqai", "groq", "togetherai", "together"]:
        if not session.get("llm_enabled"):
            session["llm_provider"] = "in-progress"
            chat_history = prompt_for_llm_provider(chat_history)
            session["chat_history"] = chat_history
            
            return jsonify({"chat_history": chat_history[-1:]})
            
        elif session.get("llm_provider") and len(session.get("valid_providers", {})) > 0:
            if session.get("llm_provider") == "in-progress":
                # Process the LLM provider selection
                updated_history, needs_api_key = handle_llm_provider_selection(user_msg, chat_history)
                
                # Update the chat history
                chat_history = updated_history
                session["chat_history"] = chat_history
                
                # If we need an API key, indicate that to the frontend
                if needs_api_key:
                    return jsonify({
                        "chat_history": chat_history[-1:],
                        "require_api_key": True,
                        "provider": session.get("llm_provider")
                    })
                
                return jsonify({"chat_history": chat_history[-1:]})
                
            else:
                # Provider already selected, check if we need API key
                if not session.get("api_enabled"):
                    return jsonify({
                        "require_api_key": True, 
                        "provider": session.get("llm_provider"),
                        "chat_history": [{
                            "sender": "bot",
                            "text": f"Please provide your {session.get('llm_provider')} API key to continue."
                        }]
                    })
                
                # If we have API key, generate explanation
                try:
                    # Get relevant data for explanation
                    table_id = session.get("table_id", 0)
                    all_plots = session.get("plots")

                    table_plot = [item for item in all_plots if item.get("id") == table_id][0]
                    table_plot_html = table_plot["srcdoc"]

                    providers = session.get("valid_providers")
                    selected_model = session.get("llm_provider")
                    llm = providers[selected_model]()
                    template = [
                        ("system", "You are a fairness analysis expert, specially skilled to point out trade-offs between different performance metrics and fairness metrics, and gives suggestive guidance to user where fairness needs to be improved."),
                        ("user", "Please explain in brief summary the following given report focusing mainly on the most important numbers about fairness and predictive performance across protected attributes:\n{context}\n\n")
                    ]
                    llm_agent = get_langchain_agent(llm, template)
                    summary_response = summarize_html_report(table_plot_html, llm_agent)
                    
                except Exception as e:
                    if "openai" in str(e).lower() or "key" in str(e).lower() or "quota" in str(e).lower():
                        return jsonify({
                            "require_api_key": True, 
                            "provider": session.get("llm_provider"),
                            "chat_history": [{
                                "sender": "bot",
                                "text": f"There was an issue with your API key. Please provide a valid {session.get('llm_provider')} API key."
                            }]
                        })
                
                    summary_response = f"(⚠️ Could not summarize the plot: {e})"
            
                chat_history.append({"sender": "bot", "text": summary_response})
                # Clean up session
                if "valid_providers" in session:
                    del session["valid_providers"]
                session["last_action"] = "summarized"
                session["chat_history"] = chat_history
                return jsonify({"chat_history": chat_history[-1:]})
        
        elif session.get("llm_provider") == "in-progress" and len(session.get("valid_providers", {})) == 0:
            chat_history.append({
                "sender": "bot",
                "text": "You can also choose a Theta index below to update your model, and corresponding model's performance will also be updated. To start again with another Data/ Model, click on 'reset chat' 🙂"
            })
            return jsonify({"chat_history": chat_history[-1:]})
        
    # Handle non-protected group selection
    if user_msg.startswith("nprotg_") and nprotgs_value:
        # Initialize nprotgs if not exists
        if "nprotgs_temp" not in session:
            session["nprotgs_temp"] = []

        # Add the new value
        session["nprotgs_temp"].append(nprotgs_value)

        # Check if we have enough values
        if len(session["nprotgs_temp"]) >= len(user_args.get("prots", [])):
            # We have all values, combine them
            user_args["nprotgs"] = session["nprotgs_temp"]
            chat_history.append(
                {
                    "sender": "bot",
                    "text": f"Set nprotgs = {' '.join(session['nprotgs_temp'])}",
                }
            )

            # Clear temporary storage
            session.pop("nprotgs_temp", None)

            # Continue to next parameter
            new_missing = get_missing_args(user_args)
            if new_missing:
                next_arg = new_missing[0]
                prompt, options = get_prompt_for_arg(next_arg, user_args)
                chat_history.append(
                    {"sender": "bot", "text": prompt, "options": options}
                )
            else:
                chat_history.append(
                    {
                        "sender": "bot",
                        "text": "All arguments captured. What would you like to do?",
                        "options": [
                            {"value": "run", "text": "Run Training"},
                            {"value": "reset", "text": "Start Over"},
                        ],
                    }
                )

            session["chat_history"] = chat_history
            session["user_args"] = user_args
            return jsonify({"chat_history": chat_history[-2:]})
        else:
            # Still need more values, show progress
            remaining = len(user_args.get("prots", [])) - len(session["nprotgs_temp"])
            protected_attrs = user_args.get("prots", [])
            chat_history.append(
                {
                    "sender": "bot",
                    "text": f"Selected {nprotgs_value} for {protected_attrs[len(session['nprotgs_temp'])-1]}. "
                    + f"Please select {remaining} more value(s) for: {', '.join(protected_attrs[len(session['nprotgs_temp']):])}",
                }
            )

            # Show options for the next protected attribute
            next_prot = protected_attrs[len(session["nprotgs_temp"])]
            dataset_name = user_args.get("dataset", "").lower()

            if (
                dataset_name in NONPROTECTED_SUGGESTIONS
                and next_prot in NONPROTECTED_SUGGESTIONS[dataset_name]
            ):
                options = NONPROTECTED_SUGGESTIONS[dataset_name][next_prot]
                chat_history.append(
                    {
                        "sender": "bot",
                        "text": f"Select non-protected value for {next_prot}:",
                        "options": [
                            {
                                "value": f"nprotg_{next_prot}_{option['value']}",
                                "text": option["text"],
                            }
                            for option in options
                        ],
                    }
                )
            elif dataset_name == "upload_data":
                    options = [
                        {"value": str(val), "text": str(val)}
                        for val in sorted(set(session["data"].data[next_prot]))
                    ]
                    chat_history.append(
                        {
                            "sender": "bot",
                            "text": f"Select non-protected value for {next_prot}:",
                            "options": [
                                {
                                    "value": f"nprotg_{next_prot}_{val['value']}",
                                    "text": val["text"],
                                }
                                for val in options
                            ],
                        }
                    )

            session["chat_history"] = chat_history
            return jsonify({"chat_history": chat_history[-2:]})
            

    # Handle feature selection submission
    if user_msg == "submit_features" and selected_features:
        # Process submitted features
        if not selected_features:
            chat_history.append(
                {
                    "sender": "bot",
                    "text": "Error: Please select at least one protected attribute.",
                }
            )
        else:
            user_args["prots"] = selected_features
            chat_history.append(
                {
                    "sender": "bot",
                    "text": f"Set protected attributes = {', '.join(selected_features)}",
                }
            )

            # Continue to next parameter
            new_missing = get_missing_args(user_args)
            if new_missing:
                next_arg = new_missing[0]
                prompt, options = get_prompt_for_arg(next_arg, user_args)

                message = {"sender": "bot", "text": prompt}
                if options:
                    message["options"] = options

                chat_history.append(message)
            else:
                chat_history.append(
                    {
                        "sender": "bot",
                        "text": "All arguments captured. What would you like to do?",
                        "options": [
                            {"value": "run", "text": "Run Training"},
                            {"value": "reset", "text": "Start Over"},
                        ],
                    }
                )

            session["chat_history"] = chat_history
            session["user_args"] = user_args
            return jsonify({"chat_history": chat_history[-2:]})

    if user_msg == "visualize_yes":
        # Store in user_args so we don't ask again
        user_args["data_visualization"] = "show_visualization"
        session["user_args"] = user_args
        # User wants to visualize data
        return jsonify({"redirect": "/visualize_data"})
    elif user_msg == "visualize_no":
        # Store in user_args so we don't ask again
        user_args["data_visualization"] = "skip_visualization"
        session["user_args"] = user_args
        # User doesn't want visualization, proceed to next step
        new_missing = get_missing_args(user_args)
        if new_missing:
            next_arg = new_missing[0]
            prompt, options = get_prompt_for_arg(next_arg, user_args)

            message = {"sender": "bot", "text": prompt}
            if options:
                message["options"] = options

            chat_history.append(message)
        else:
            chat_history.append(
                {
                    "sender": "bot",
                    "text": "All arguments captured. What would you like to do?",
                    "options": [
                        {"value": "run", "text": "Run Training"},
                        {"value": "reset", "text": "Start Over"},
                    ],
                }
            )

        session["chat_history"] = chat_history
        return jsonify({"chat_history": chat_history[-1:]})

    # **NEW FEATURE**: If user chooses "default", skip everything and run with default_args
    if not user_args and user_msg.lower() == "default":
        for arg in default_args:
            if arg not in user_args:
                user_args[arg] = default_args[arg]
        session["classifier"] = user_args["classifier"]
        chat_history.append(
            {"sender": "bot", "text": "Running MMM_Fair with default parameters..."}
        )
        visualizations, html_divs = run_mmm_fair_app(default_args)

        session["last_action"] = "plotted"

        default_resp="✅ Training complete! Here in the generated pareto plots (upper box) you see various solution points (models) available, where each solution point (denoted by theta: number) shows a different trade-off point between the different training objectives. \n\n The performance of the suggested best model is also plotted (lower box). You can also choose a Theta index below to update your model, and corresponding model's performance will also be updated."

        return jsonify(
            {
                "chat_history": [
                    {
                        "sender": "bot",
                        "text": f"{default_resp}\n\n If you'd like me to explain what you're seeing?.\n",
                        "options": [
                            {"value": "explain", "text": "Yes, please explain"},
                            {"value": "reset", "text": "Start Over"},
                        ],
                    }
                ],
                "plots": visualizations,
                "html_divs": html_divs,
            }
        )

    # Add user message
    if user_msg:
        # For button clicks, show the button text instead of value
        display_text = user_msg
        if chat_history:  # Check if chat_history is not empty
            options = chat_history[-1].get("options")
            if isinstance(options, list):
                for option in options: #chat_history[-1].get("options", []):
                    if option["value"] == user_msg:
                        display_text = option["text"]
                        break

        chat_history.append({"sender": "user", "text": display_text})

    missing_args = get_missing_args(user_args)

    if not missing_args:
        if user_msg.lower() == "run":
            for arg in default_args:
                if arg not in user_args:
                    user_args[arg] = default_args[arg]
            session["classifier"] = user_args["classifier"]
            visualizations, html_divs = run_mmm_fair_app(user_args)

            session["last_action"] = "plotted"
            default_resp="✅ Training complete! Here in the generated pareto plots (upper box) you see various solution points (models) available, where each solution point (denoted by theta: number) shows a different trade-off point between the different training objectives. \n\n The performance of the suggested best model is also plotted (lower box). You can also choose a Theta index below to update your model, and corresponding model's performance will also be updated."

            return jsonify(
                {
                    "chat_history": [
                        {
                            "sender": "bot",
                            "text": f"{default_resp}\n\n If you'd like me to explain what you're seeing?.\n",
                            "options": [
                                {"value": "explain", "text": "Yes, please explain"},
                                {"value": "reset", "text": "Start Over"},
                            ],
                        }
                    ],
                    "plots": visualizations,
                    "html_divs": html_divs,

                }
            )
        else:
            chat_history.append(
                {
                    "sender": "bot",
                    "text": "We have all arguments. What would you like to do?",
                    "options": [
                        {"value": "run", "text": "Run Training"},
                        {"value": "reset", "text": "Start Over"},
                    ],
                }
            )
    else:
        current_arg = missing_args[0]
        print(missing_args)
        valid, clean_val, err_msg = validate_arg(current_arg, user_msg, user_args)

        # Process classifier selection (from buttons)
        if current_arg == "classifier":
            if user_msg in ["MMM_Fair", "MMM_Fair_GBT"]:
                user_args[current_arg] = user_msg
                valid = True
                clean_val = user_msg

        # **NEW FEATURE**: Load dataset after fetching "dataset" argument
        if current_arg == "dataset":
            #print("we entered here")
            if valid:
                # Load dataset and store for later use
                try:
                    if "file" in request.files:
                        uploaded_file = request.files.get("file")
                        #print(f" - Filename: {uploaded_file.filename}")
                        df = pd.read_csv(uploaded_file)
                        session["data"] = data_local(df)
                        user_args["dataset"] = "upload_data"
                    else:    
                        session["data"] = data_uci(clean_val)  # Load dataset
                        user_args[current_arg] = clean_val
                    session["user_args"] = user_args#clean_val
                    chat_history.append(
                        {"sender": "bot", "text": f"Set {current_arg} = {clean_val}."}
                    )
                    chat_history.append(
                        {"sender": "bot", "text": "Dataset loaded successfully."}
                    )

                except Exception as e:
                    chat_history.append(
                        {
                            "sender": "bot",
                            "text": f"Error loading dataset: {str(e)}\nPlease select a valid dataset.",
                            "options": [
                                {"value": "adult", "text": "Adult Dataset"},
                                {"value": "bank", "text": "Bank Dataset"},
                                {"value": "credit", "text": "Credit Dataset"},
                                {"value": "kdd", "text": "KDD Dataset"},
                                {"value": "upload_data", "text": "📁 Upload your own Data (currently supported files: '.csv'"},
                            ],
                        }
                    )
                    session["chat_history"] = chat_history
                    return jsonify({"chat_history": chat_history[-2:]})

            else:
                chat_history.append(
                    {
                        "sender": "bot",
                        "text": f"Error: {err_msg}\nPlease select a dataset:",
                        "options": [
                            {"value": "adult", "text": "Adult Dataset"},
                            {"value": "bank", "text": "Bank Dataset"},
                            {"value": "credit", "text": "Credit Dataset"},
                            {"value": "kdd", "text": "KDD Dataset"},
                            {"value": "upload_data", "text": "📁 Upload your own Data (currently supported files: '.csv'"},
                        ],
                    }
                )
                session["chat_history"] = chat_history
                return jsonify({"chat_history": chat_history[-2:]})

        # Special handling for protected attributes
        elif current_arg == "prots":
            # Instead of validating text input, we'll create the feature selector UI
            prompt, options = get_prompt_for_arg(current_arg, user_args)

            message = {"sender": "bot", "text": prompt}

            if options:
                message["options"] = options

            chat_history.append(message)
            session["chat_history"] = chat_history
            return jsonify({"chat_history": chat_history[-1:]})

        elif current_arg == "constraint":
            if not valid:
                chat_history.append(
                    {
                        "sender": "bot",
                        "text": f"Error: {err_msg}\nPlease select a fairness constraint:",
                        "options": [
                            {"value": "DP", "text": "Demographic Parity (DP)"},
                            {"value": "EP", "text": "Equal Precision (EP)"},
                            {"value": "EO", "text": "Equal Opportunity (EO)"},
                            {"value": "TPR", "text": "True Positive Rate (TPR)"},
                            {"value": "FPR", "text": "False Positive Rate (FPR)"},
                        ],
                    }
                )
                session["chat_history"] = chat_history
                return jsonify({"chat_history": chat_history[-2:]})
            else:
                user_args[current_arg] = clean_val
                chat_history.append(
                    {"sender": "bot", "text": f"Set {current_arg} = {clean_val}."}
                )

        elif current_arg == "data_visualization" and valid:
            user_args[current_arg] = clean_val

            if clean_val == "show_visualization":
                # Redirect to the visualization endpoint
                return jsonify({"redirect": "/visualize_data"})
            else:
                # Just set the value and continue to next argument
                chat_history.append(
                    {"sender": "bot", "text": "Continuing without data visualization."}
                )

                # Now process as usual for the next missing argument
                new_missing = get_missing_args(user_args)
                if new_missing:
                    next_arg = new_missing[0]
                    prompt, options = get_prompt_for_arg(next_arg, user_args)

                    message = {"sender": "bot", "text": prompt}
                    if options:
                        message["options"] = options

                    chat_history.append(message)
                else:
                    chat_history.append(
                        {
                            "sender": "bot",
                            "text": "All arguments captured. What would you like to do?",
                            "options": [
                                {"value": "run", "text": "Run Training"},
                                {"value": "reset", "text": "Start Over"},
                            ],
                        }
                    )

                session["chat_history"] = chat_history
                session["user_args"] = user_args
                return jsonify({"chat_history": chat_history[-2:]})

        else:
            # For other arguments, still process text input but provide buttons when possible
            if not valid:
                chat_history.append(
                    {
                        "sender": "bot",
                        "text": f"Error: {err_msg}\nPlease re-enter {current_arg}.",
                    }
                )
            else:
                user_args[current_arg] = clean_val
                chat_history.append(
                    {"sender": "bot", "text": f"Set {current_arg} = {clean_val}."}
                )

        new_missing = get_missing_args(user_args)
        if new_missing:
            next_arg = new_missing[0]
            prompt, options = get_prompt_for_arg(next_arg, user_args)

            message = {"sender": "bot", "text": prompt}
            if options:
                message["options"] = options

            chat_history.append(message)
        else:
            chat_history.append(
                {
                    "sender": "bot",
                    "text": "All arguments captured. What would you like to do?",
                    "options": [
                        {"value": "run", "text": "Run Training"},
                        {"value": "reset", "text": "Start Over"},
                    ],
                }
            )

    session["chat_history"] = chat_history
    session["user_args"] = user_args
    return jsonify({"chat_history": chat_history[-2:]})


@app.route("/update_model", methods=["POST"])
def update_model():
    # Get Theta value from the request
    data = request.json
    theta_value = int(data.get("theta", -1))

    # Retrieve trained classifier
    mmm_classifier = session.get("mmm_classifier")
    if not mmm_classifier:
        return jsonify(
            {"success": False, "error": "No trained model found! Run training first."}
        )

    # Validate Theta index
    if theta_value < 0 or theta_value >= len(mmm_classifier.ob):
        return jsonify(
            {
                "success": False,
                "error": f"Invalid Theta index. Please select between 0 and {len(mmm_classifier.ob) - 1}.",
            }
        )

    # Update the model with selected Theta
    mmm_classifier.update_theta(theta=theta_value)
    session["mmm_classifier"] = mmm_classifier  # Store updated model

    X_test = session.get("xtest")
    y_test = session.get("ytest")
    user_args = session.get("user_args", {})
    sensitives = session.get("sensitives")
    saIndex_test = session.get("saIndex_test")

    y_pred = mmm_classifier.predict(X_test)

    report_table = generate_reports(
        "html", sensitives, mmm_classifier, saIndex_test, y_pred, y_test, html=True
    )

    table_id = session.get("table_id", 0)

    print(f"DEBUG: Updating report table with ID: {table_id}")

    report_table_plot_dict = {
        "updated_data":
        {
            "srcdoc": report_table,
        },
        "existing_id": table_id
    }


    # unique_id = str(uuid.uuid4())[:4]
    # plot_table = "table_.html"
    # p2 = f"table_{unique_id}.html"
    # plot_table_path = os.path.join(PLOT_DIR, p2)  # plot_table)

    # with open(plot_table_path, "w") as f:
    #     f.write(report_table)
    #     f.flush()
    #     os.fsync(f.fileno())

    # session["plot_fair_url"] = f"/static/{p2}"
    return jsonify(
        {
            "success": True,
            "message": f"Model updated with Theta index {theta_value}.",
            # "plot_fair_url": f"/static/{plot_table}",
            "update_plots": [report_table_plot_dict],
            "chat_history": [
                {
                    "sender": "bot",
                    "text": f"Model updated with Theta index {theta_value}. Would you like to save this model?",
                    "options": [
                        {"value": "save_model", "text": "Save Model"},
                        {"value": "reset", "text": "Start Over"},
                    ],
                }
            ],
        }
    )


def get_missing_args(user_args):
    """
    Return the list of arguments we still need, considering
    if user chose MMM_Fair_GBT => skip base_learner
    """
    # chosen_data = user_args.get("data", "").lower()
    # if not chosen_data:
    #     return ["dataset"]

    needed_args = []
    for arg in CHAT_ARGS:
        if arg == "base_learner":
            chosen_classifier = user_args.get("classifier", "").lower()
            if not chosen_classifier:
                return ["classifier"]
            elif chosen_classifier in [
            "mmm_fair_gbt",
            "mmm-fair-gbt",
            ]:
                continue  # Skip base_learner for GBT models

        if arg == "data_visualization":
            has_target = "target" in user_args and user_args["target"]
            has_prots = "prots" in user_args and user_args["prots"]

            if has_target and has_prots and "data_visualization" not in user_args:
                needed_args.append(arg)
            continue  # Skip to next argument regardless

        if arg not in user_args:
            needed_args.append(arg)

    return needed_args


def validate_arg(arg_name, user_input, user_args):
    """
    Simple validator that can parse or store defaults.
    Returns (valid, clean_value, error_message).
    """
    print(arg_name, user_input, user_args)
    if arg_name == "dataset":
        if user_input.lower().endswith(".csv"):
            return True, user_input, ""
        elif user_input.lower() in ["adult", "bank", "kdd", "credit","upload_data"]:
            return True, user_input.lower(), "Loading Data...please wait!!!"
        
        else:
            return False, user_input, ""  # Fallback case
    

    elif arg_name == "n_learners":
        if not user_input.isdigit():
            return False, None, "n_learners must be an integer e.g. 100"
        return True, int(user_input), ""

    elif arg_name == "constraint":
        c = user_input.upper()
        if c not in ["DP", "EP", "EO", "TPR", "FPR"]:
            return False, None, "Constraint must be DP, EP, EO, TPR, or FPR."
        return True, c, ""

    elif arg_name == "classifier":
        val = user_input.lower()
        if val not in ["mmm_fair", "mmm_fair_gbt", "mmm-fair", "mmm-fair-gbt"]:
            return False, None, "Classifier must be 'MMM_Fair' or 'MMM_Fair_GBT'."
        return True, "MMM_Fair_GBT" if "gbt" in val else "MMM_Fair", ""

    elif arg_name == "target":
        # Extract the name of the target column
        available_target_name = (
            session["data"].labels["label"].name
        )  # Get the column name

        #print(
        #    f"DEBUG: Available target name: {available_target_name}"
        #)  # Debugging output

        return True, user_input, ""

    elif arg_name == "data_visualization":
        if user_input.lower() in ["visualize_yes", "yes", "y"]:
            # Instead of storing a simple value, store a flag to redirect
            # This will be handled specially in ask_chat
            return True, "show_visualization", ""
        else:
            return True, "skip_visualization", ""

    elif arg_name == "pos_Class":
        if not user_input:
            return True, None, ""  # Default None
        return True, user_input, ""

    elif arg_name == "prots":
        available_columns = session["data"].data.columns.tolist()
        available_columns = [v.lower() for v in available_columns]
        if not user_input:
            return False, [], "At least one protected attribute expected"
        else:
            k = ""
            for prot in user_input.split(" "):
                if prot.lower() not in available_columns:
                    chosen_dataset = user_args.get("dataset", "").lower()
                    if chosen_dataset not in ["adult", "bank"]:
                        return (
                            False,
                            None,
                            f"Invalid protected. Available list of attributes in the data: {available_columns}.",
                        )
                    else:
                        k += " " + prot + ","

            if k != "":
                return (
                    True,
                    user_input.split(),
                    f"Invalid protected attribute(s) {k} will be replaced with default known protected.",
                )

            return True, user_input.split(), ""

    elif arg_name == "nprotgs":
        # If we're validating from button selection, we already handled the count check
        if "nprotgs_temp" in session and len(session["nprotgs_temp"]) > 0:
            return True, user_input.split(), ""
    
        # Otherwise, default validation
        existing_prots = user_args.get("prots", [])
        splitted = user_input.split()
        if len(splitted) != len(existing_prots):
            return (
                False,
                None,
                f"Got {len(splitted)} non-protected vals for {len(existing_prots)} protected columns. Please match count.",
            )
    
        # ✅ Add this return for the normal case
        return True, splitted, ""
    #elif arg_name == "nprotgs":
        # # If we're validating from button selection, we already handled the count check
        # if "nprotgs_temp" in session and len(session["nprotgs_temp"]) > 0:
        #     return True, user_input.split(), ""

        # # Otherwise, default validation
        # existing_prots = user_args.get("prots", [])
        # splitted = user_input.split()
        # if len(splitted) != len(existing_prots):
        #     return (
        #         False,
        #         None,
        #         f"Got {len(splitted)} non-protected vals for {len(existing_prots)} protected columns. Please match count.",
        #     )

    elif arg_name == "deploy":
        val = user_input.lower()
        if val not in ["onnx", "pickle"]:
            return False, None, "Deployment must be 'onnx' or 'pickle'."
        return True, val, ""

    elif arg_name == "moo_vis":
        if user_input.lower() in ["true", "yes", "y"]:
            return True, True, ""
        else:
            return True, False, ""

    else:
        return True, user_input, ""  # Default case


# Modified get_prompt_for_arg to provide dataset features
def get_prompt_for_arg(arg_name, user_args):
    """
    Return a question/prompt for the user and available button options if applicable
    based on arg_name
    """
    prompts = {
        "dataset": (
            "Please select a dataset:",
            [
                {"value": "adult", "text": "Adult Dataset"},
                {"value": "bank", "text": "Bank Dataset"},
                {"value": "credit", "text": "Credit Dataset"},
                {"value": "kdd", "text": "KDD Dataset"},
            ],
        ),
        "target": (
            "Enter the label (target) column in your dataset (e.g. 'income'):",
            None,
        ),
        "pos_Class": (
            "Enter the positive class label if known (else press Enter to skip):",
            None,
        ),
        "n_learners": (
            "How many learners / iterations?",
            [
                {"value": "50", "text": "50 Learners"},
                {"value": "100", "text": "100 Learners"},
                {"value": "200", "text": "200 Learners"},
            ],
        ),
        "constraint": (
            "Please select a fairness constraint:",
            [
                {"value": "DP", "text": "Demographic Parity (DP)"},
                {"value": "EP", "text": "Equal Precision (EP)"},
                {"value": "EO", "text": "Equal Opportunity (EO)"},
                {"value": "TPR", "text": "True Positive Rate (TPR)"},
                {"value": "FPR", "text": "False Positive Rate (FPR)"},
            ],
        ),
        "nprotgs": (
            "Enter corresponding non-protected spec, e.g. 'White Male 30_60' matching the above columns:",
            None,
        ),
        "deploy": (
            "Would you like to deploy as 'onnx' or 'pickle'?",
            [
                {"value": "onnx", "text": "ONNX Format"},
                {"value": "pickle", "text": "Pickle Format"},
            ],
        ),
        "moo_vis": (
            "Do you want to enable multi-objective visualization?",
            [
                {"value": "True", "text": "Yes, enable visualization"},
                {"value": "False", "text": "No, disable visualization"},
            ],
        ),
        "data_visualization": (
            "Would you like to visualize the distribution of your data?",
            [
                {"value": "visualize_yes", "text": "Yes, show visualization"},
                {"value": "visualize_no", "text": "No, continue without visualization"},
            ],
        ),
    }

    if arg_name == "classifier":
        return (
            "Please select a model type:",
            [
                {"value": "MMM_Fair", "text": "MMM_Fair (AdaBoost)"},
                {"value": "MMM_Fair_GBT", "text": "MMM_Fair_GBT (Gradient Boosting)"},
            ],
        )

    if arg_name == "dataset":
        return  (
            "Please select a dataset:",
            [
                {"value": "adult", "text": "Adult Dataset"},
                {"value": "bank", "text": "Bank Dataset"},
                {"value": "credit", "text": "Credit Dataset"},
                {"value": "kdd", "text": "KDD Dataset"},
                {"value": "upload_data", "text": "📁 Upload your own Data (currently supported types: '.csv'"},
            {"value": "default", "text": "Run with default setup on Adult data"}
            ],
        )
        
    if arg_name == "base_learner":
        return (
            "Select a base learner:",
            [
                {"value": "tree", "text": "Decision Tree"},
                {"value": "lr", "text": "Linear Regression"},
                {"value": "logistic", "text": "Logistic Regression"},
                {"value": "extratree", "text": "Extra Tree"},
            ],
        )

    if arg_name == "prots":
        # Get dataset name
        dataset_name = user_args.get("dataset", "").lower()
        #print(f'Debugging inside get_prompt {dataset_name}')
        available_features = []
        recommended_features = []

        if "data" in session:
            #print("we came here for data")
            # Get actual columns from the dataset
            available_features = session["data"].data.columns.tolist()
            #print(f'Debugging inside get_prompt features {available_features}')
            DATASET_FEATURES[dataset_name] = available_features
        elif dataset_name in DATASET_FEATURES:
            # Use cached features if available
            available_features = DATASET_FEATURES[dataset_name]

        # Get recommended features for this dataset
        #recommended_features = []
        if dataset_name in DATASET_RECOMMENDATIONS:
            recommended_features = DATASET_RECOMMENDATIONS[dataset_name]

        #if dataset_name == "upload_data":
        #    recomended_features= available_features
        # Build a special message with recommendations
        message = "Please select protected attributes:"
        if recommended_features:
            message += f"\n\nRecommended for {dataset_name.capitalize()}: " + ", ".join(
                recommended_features
            )

        # Return a special format with available features
        
        print(f'Debugging for recomended values {recommended_features}')
        return (
            message,
            {
                "type": "features_selector",
                "available": available_features,
                "recommended": recommended_features,
            },
        )

    elif arg_name == "nprotgs":
        dataset_name = user_args.get("dataset", "").lower()
        protected_attrs = user_args.get("prots", [])
    
        # Case 1: Use default suggestions for known datasets
        if dataset_name in NONPROTECTED_SUGGESTIONS and protected_attrs:
            if "nprotgs_temp" in session and session["nprotgs_temp"]:
                next_idx = len(session["nprotgs_temp"])
                if next_idx < len(protected_attrs):
                    next_prot = protected_attrs[next_idx]
                    if next_prot in NONPROTECTED_SUGGESTIONS[dataset_name]:
                        options = NONPROTECTED_SUGGESTIONS[dataset_name][next_prot]
                        return (
                            f"Select non-protected value for {next_prot}:",
                            [
                                {
                                    "value": f"nprotg_{next_prot}_{option['value']}",
                                    "text": option["text"],
                                }
                                for option in options
                            ],
                        )
            elif protected_attrs:
                first_prot = protected_attrs[0]
                if first_prot in NONPROTECTED_SUGGESTIONS[dataset_name]:
                    options = NONPROTECTED_SUGGESTIONS[dataset_name][first_prot]
                    return (
                        f"Select non-protected value for {first_prot}:",
                        [
                            {
                                "value": f"nprotg_{first_prot}_{option['value']}",
                                "text": option["text"],
                            }
                            for option in options
                        ],
                    )
    
        # Case 2: Dynamically build dropdown from uploaded CSV
        if dataset_name == "upload_data" and "data" in session:
            df = session["data"].data
            if protected_attrs:
                temp_vals = session.get("nprotgs_temp", [])
                next_idx = len(temp_vals)
                if next_idx < len(protected_attrs):
                    next_prot = protected_attrs[next_idx]
                    try:
                        values = sorted(set(df[next_prot].dropna().unique()))
                        options = [
                            {
                                "value": f"nprotg_{next_prot}_{val}",
                                "text": str(val),
                            }
                            for val in values
                        ]
                        return (
                            f"Select non-protected value for '{next_prot}':",
                            options
                        )
                    except Exception as e:
                        return (
                            f"⚠️ Could not retrieve values for '{next_prot}': {e}",
                            None
                        )
            else:
                session["nprotgs_temp"] = []
    
        return (
            "Enter corresponding non-protected spec, e.g. 'White Male' matching the above columns:",
            None,
        )
    elif arg_name == "target":
        target=session['data'].labels['label'].name
        dataset_name = user_args.get("dataset", "").lower()
        if dataset_name in TARGET_SUGGESTIONS:
            if dataset_name == "upload_data":
                TARGET_SUGGESTIONS[dataset_name]= [{"value": target, "text": target},]
            return (
                "Select the label (target) column in your dataset:",
                TARGET_SUGGESTIONS[dataset_name],
            )
        else:
            target=session['data'].labels['label'].name
            print(f'Debugging target name{target}')
            return (
                "Enter the label (target) column in your dataset:",
                {"value": target, "text": target},
            )

    elif arg_name == "pos_Class":
        dataset_name = user_args.get("dataset", "").lower()
        target_column = user_args.get("target", "")
    
        try:
            if "data" in session and target_column:
                labels = session["data"].labels["label"]
                unique_classes = list(set(labels))
                options = [{"value": str(c), "text": f"Use '{c}' as positive"} for c in unique_classes]
    
                # Add a "Skip" option too
                options.insert(0, {"value": "", "text": "Skip (use default)"})
                return ("Please select the positive class label:", options)
    
        except Exception as e:
            print(f"DEBUG: Failed to fetch pos_Class options — {e}")
    
        return ("Enter the positive class label if known (else press Enter to skip):", None)

    return prompts.get(arg_name, (f"Enter {arg_name}:", None))


# Add a new route to handle feature selection
@app.route("/select_feature", methods=["POST"])
def select_feature():
    data = request.json
    feature = data.get("feature", "")

    # Get current selected features
    user_args = session.get("user_args", {})
    current_prots = user_args.get("prots", [])

    # Add the feature if not already selected
    if feature and feature not in current_prots:
        current_prots.append(feature)
        user_args["prots"] = current_prots
        session["user_args"] = user_args

    return jsonify({"success": True, "selected_features": current_prots})


@app.route("/visualize_data", methods=["POST"])
def visualize_data():
    """Generate and return a visualization of the dataset distribution"""
    # Get data from session
    if "data" not in session:
        return jsonify(
            {
                "success": False,
                "error": "No dataset loaded. Please load a dataset first.",
            }
        )

    data_obj = session.get("data")
    user_args = session.get("user_args", {})

    # Get target and protected attributes
    target = user_args.get("target")
    protected_attrs = user_args.get("prots", [])

    if not target or not protected_attrs:
        return jsonify(
            {"success": False, "error": "Target or protected attributes not defined."}
        )

    try:
        # Generate the plot using your existing function
        df = data_obj.data.copy()  # Get the dataframe from your data object
        df[target] = data_obj.labels["label"].values.copy()
        # Generate a unique filename
        # unique_id = str(uuid.uuid4())[:8]
        # pie_plot_filename = f"data_dist_{unique_id}.html"
        # pie_plot_path = os.path.join(PLOT_DIR, pie_plot_filename)

        # Generate the nested pie chart
        data_plot_html = generate_nested_pie_chart(df, [target] + protected_attrs)

        
        style_dict = {  
            "width": "100%",
            "height": "500px",
            "overflow": "hidden",
            "border": "None",     
        }

        data_plot_dict = {
            "srcdoc": data_plot_html,
            "style": style_dict,
            "id": str(uuid.uuid4())[:4]
        }

        # print(plot_html)
        # # Save the plot
        # with open(pie_plot_path, "w") as f:
        #     f.write(plot_html)
        #     f.flush()
        #     os.fsync(f.fileno())
        #     # Get the next prompt to display after visualization

        # Mark visualization as completed in user_args
        user_args["data_visualization"] = "completed"
        session["user_args"] = user_args

        # Get the next prompt to display after visualization
        chat_history = session.get("chat_history", [])
        chat_history.append(
            {
                "sender": "bot",
                "text": "Here's the distribution of your data based on target and protected attributes.",
            }
        )

        # Determine the next argument to prompt for
        new_missing = get_missing_args(user_args)

        if new_missing:
            next_arg = new_missing[0]
            prompt, options = get_prompt_for_arg(next_arg, user_args)

            next_message = {"sender": "bot", "text": prompt}
            if options:
                next_message["options"] = options

            chat_history.append(next_message)
        else:
            # If no more arguments needed, show the run options
            chat_history.append(
                {
                    "sender": "bot",
                    "text": "All arguments captured. What would you like to do?",
                    "options": [
                        {"value": "run", "text": "Run Training"},
                        {"value": "reset", "text": "Start Over"},
                    ],
                }
            )

        session["chat_history"] = chat_history

        return jsonify(
            {
                "success": True,
                "plots": [data_plot_dict],  # Send HTML directly instead of file path
                "chat_history": chat_history[
                    -2:
                ],  # Return both visualization message and next prompt
            }
        )
    except Exception as e:
        print(f"Error in visualization: {str(e)}")
        return jsonify(
            {"success": False, "error": f"Error generating visualization: {str(e)}"}
        )


@app.route("/finish_features", methods=["POST"])
def finish_features():
    user_args = session.get("user_args", {})
    selected_features = user_args.get("prots", [])

    if not selected_features:
        return jsonify(
            {
                "success": False,
                "error": "Please select at least one protected attribute.",
            }
        )

    # Convert to the format expected by the existing code
    chat_history = session.get("chat_history", [])

    # Add message showing selected features
    chat_history.append(
        {
            "sender": "bot",
            "text": f"Selected protected attributes: {', '.join(selected_features)}",
        }
    )

    # Check if we have both target and protected attributes
    if "target" in user_args and user_args["target"]:
        # Add suggestion to visualize data
        chat_history.append(
            {
                "sender": "bot",
                "text": "Would you like to visualize the distribution of your data?",
                "options": [
                    {"value": "visualize_yes", "text": "Yes, show visualization"},
                    {
                        "value": "visualize_no",
                        "text": "No, continue without visualization",
                    },
                ],
            }
        )

        session["chat_history"] = chat_history
        return jsonify({"success": True, "chat_history": chat_history[-2:]})

    # If no target yet, continue with original flow
    # Continue to next parameter
    new_missing = get_missing_args(user_args)
    if new_missing:
        next_arg = new_missing[0]
        prompt, options = get_prompt_for_arg(next_arg, user_args)

        message = {"sender": "bot", "text": prompt}
        if options:
            message["options"] = options

        chat_history.append(message)
    else:
        chat_history.append(
            {
                "sender": "bot",
                "text": "All arguments captured. What would you like to do?",
                "options": [
                    {"value": "run", "text": "Run Training"},
                    {"value": "reset", "text": "Start Over"},
                ],
            }
        )

    session["chat_history"] = chat_history
    return jsonify({"success": True, "chat_history": chat_history[-2:]})


@app.route("/reset_chat")
def reset_chat():
    session.pop("chat_history", None)
    session.pop("user_args", None)
    session.pop("data", None)
    for fname in os.listdir(PLOT_DIR):
        if (
            fname.startswith("table_")
            or fname.startswith("fair_")
            or fname.startswith("all_")
        ):
            try:
                os.remove(os.path.join(PLOT_DIR, fname))
            except Exception:
                pass
    return "Chat reset done."


def run_mmm_fair_app(user_args):
    data = session.get('data')
    user_args["df"]=data
    args = argparse.Namespace(**user_args)
    mmm_classifier, X_test, y_test, saIndex_test, sensitives = train(args)
    session["mmm_classifier"] = mmm_classifier
    session["xtest"] = X_test
    session["ytest"] = y_test
    session["saIndex_test"] = saIndex_test
    session["sensitives"] = sensitives
    PF = np.array([mmm_classifier.ob[i] for i in range(len(mmm_classifier.ob))])
    thetas = np.arange(len(mmm_classifier.ob))
    title = f"3D Scatter Plot. Showing various trade-off points between Accuracy, Balanced Accuracy, and Maximum violation of {mmm_classifier.constraint} fairness among protected attributes."

    vis_all = plot_spider(
        objectives=PF,
        theta=thetas,
        criteria="Multi",
        axis_names=["Acc.", "Balanc. Acc", "MMM-fair"],
        title=title,
        html=True,
    )
    # vis_all = plot3d(
    #     x=PF[:, 0],
    #     y=PF[:, 1],
    #     z=PF[:, 2],
    #     theta=thetas,
    #     criteria="Multi",
    #     axis_names=["Acc.", "Balanc. Acc", "MMM-fair"],
    #     title=title,
    #     html=True,
    # )
    PF = np.array(
        [mmm_classifier.fairobs[i] for i in range(len(mmm_classifier.fairobs))]
    )
    title = "3D Scatter Plot. Showing various trade-off points between maximum violation of Demopgraphic Parity, Equal Opportunity, and Equalized odds fairness for the given set of protected attributes."
    vis_fair = plot_spider(
        objectives=PF,
        theta=thetas,
        criteria="Multi-definitions",
        axis_names=["DP", "EqOpp", "EqOdd", "TPR", "FPR"],
        title=title,
        html=True,
    )

    # vis_fair = plot3d(
    #     x=PF[:, 0],
    #     y=PF[:, 1],
    #     z=PF[:, 2],
    #     theta=thetas,
    #     criteria="Multi-definitions",
    #     axis_names=["DP", "EqOpp", "EqOdd"],
    #     title=title,
    #     html=True,
    # )

    y_pred = mmm_classifier.predict(X_test)
    report_table = generate_reports(
        "html", sensitives, mmm_classifier, saIndex_test, y_pred, y_test, html=True
    )
    # plot_all = "all_.html"
    # plot_fair = "fair_.html"
    # plot_table = "table_.html"

    # plot_all_path = os.path.join(PLOT_DIR, plot_all)
    # plot_fair_path = os.path.join(PLOT_DIR, plot_fair)
    # plot_table_path = os.path.join(PLOT_DIR, plot_table)

    # # Save the Plotly-generated HTML directly to files
    # with open(plot_all_path, "w") as f:
    #     f.write(vis_all)
    # with open(plot_fair_path, "w") as f:
    #     f.write(vis_fair)
    # with open(plot_table_path, "w") as f:
    #     f.write(report_table)

    style_dict = {  
        "width": "100%",
        "height": "500px",
        "overflow": "hidden",
        "border": "None",     
    }

    vis_all_plot_dict = {
        "srcdoc": vis_all,
        "style": style_dict,
        "id": str(uuid.uuid4())[:4]
    }
    vis_fair_plot_dict = {
        "srcdoc": vis_fair,
        "style": style_dict,
        "id": str(uuid.uuid4())[:4]
    }    
    table_id = str(uuid.uuid4())[:4]
    report_table_plot_dict = {
        "srcdoc": report_table,
        "style": style_dict,
        "id": table_id
    }
    session["plots"] = [vis_all_plot_dict, vis_fair_plot_dict, report_table_plot_dict]
    session["table_id"] = table_id
    session["html_divs"] = [THETA_DIV]

    return [vis_all_plot_dict, vis_fair_plot_dict, report_table_plot_dict], [THETA_DIV]


@app.route("/static/<path:filename>")
def serve_static_files(filename):
    return send_from_directory("static", filename)


@app.route("/static/<path:filename>")
def serve_plot(filename):
    plot_dir = os.path.abspath("static/plots")  # Ensure absolute path
    return send_from_directory(plot_dir, filename)


@app.route("/save_model", methods=["POST"])
def save_model():
    data = request.json
    save_path = data.get("save_path", "").strip()

    # Retrieve trained classifier
    mmm_classifier = session.get("mmm_classifier")
    clf = session.get("classifier")
    xdata = session.get("xtest")
    user_args = session.get("user_args", {})

    if not mmm_classifier:
        return jsonify(
            {"success": False, "error": "No trained model found! Run training first."}
        )

    # Validate save path
    if not save_path or not os.path.isdir(save_path):
        return jsonify(
            {
                "success": False,
                "error": "Invalid directory. Please select a valid folder.",
            }
        )

    # Call deploy() to save the model in the user-specified directory
    try:
        user_args["save_path"] = save_path  # Update args with the user-selected path
        convert_to_onnx(mmm_classifier, save_path, xdata, clf)
        return jsonify(
            {
                "success": True,
                "message": f"Model saved in {save_path}",
                "chat_history": [
                    {
                        "sender": "bot",
                        "text": f"Model successfully saved in {save_path}. What would you like to do next?",
                        "options": [
                            {"value": "reset", "text": "Start Over"},
                            {"value": "exit", "text": "Exit"},
                        ],
                    }
                ],
            }
        )
    except Exception as e:
        return jsonify({"success": False, "error": str(e)})


def main():
    app.run(debug=True)


if __name__ == "__main__":
    main()
